import os
import sys
import time
from unittest import mock

import pytest

import sentry_sdk
from sentry_sdk._span_batcher import SpanBatcher


def test_envelope_by_trace_id(sentry_init, capture_envelopes, monkeypatch):
    """Envelopes only contain spans of one trace ID."""
    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    envelopes = capture_envelopes()

    sentry_sdk.traces.new_trace()

    with sentry_sdk.traces.start_span(name="span 1a") as span1:
        trace_id1 = span1.trace_id
    with sentry_sdk.traces.start_span(name="span 1b"):
        pass

    sentry_sdk.traces.new_trace()

    with sentry_sdk.traces.start_span(name="span 2a") as span2:
        trace_id2 = span2.trace_id
    with sentry_sdk.traces.start_span(name="span 2b"):
        pass

    sentry_sdk.flush()

    assert len(envelopes) == 2

    assert envelopes[0].headers["trace"]["trace_id"] == trace_id1
    assert len(envelopes[0].items[0].payload.json["items"]) == 2
    assert envelopes[0].items[0].payload.json["items"][0]["name"] == "span 1a"
    assert envelopes[0].items[0].payload.json["items"][0]["trace_id"] == trace_id1
    assert envelopes[0].items[0].payload.json["items"][1]["name"] == "span 1b"
    assert envelopes[0].items[0].payload.json["items"][1]["trace_id"] == trace_id1

    assert envelopes[1].headers["trace"]["trace_id"] == trace_id2
    assert len(envelopes[1].items[0].payload.json["items"]) == 2
    assert envelopes[1].items[0].payload.json["items"][0]["name"] == "span 2a"
    assert envelopes[1].items[0].payload.json["items"][0]["trace_id"] == trace_id2
    assert envelopes[1].items[0].payload.json["items"][1]["name"] == "span 2b"
    assert envelopes[1].items[0].payload.json["items"][1]["trace_id"] == trace_id2


def test_max_envelope_size(sentry_init, capture_envelopes, monkeypatch):
    """Envelope max size is respected."""
    monkeypatch.setattr(SpanBatcher, "MAX_ENVELOPE_SIZE", 2)
    # set the time-based flush limit to something huge so that we're not observing
    # time-based flushes
    monkeypatch.setattr(SpanBatcher, "FLUSH_WAIT_TIME", 100000)

    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    envelopes = capture_envelopes()

    with sentry_sdk.traces.start_span(name="span 1"):
        pass
    with sentry_sdk.traces.start_span(name="span 2"):
        pass
    with sentry_sdk.traces.start_span(name="span 3"):
        pass
    with sentry_sdk.traces.start_span(name="span 4"):
        pass
    with sentry_sdk.traces.start_span(name="span 5"):
        pass

    sentry_sdk.flush()

    assert len(envelopes) == 3

    assert len(envelopes[0].items[0].payload.json["items"]) == 2
    assert envelopes[0].items[0].payload.json["items"][0]["name"] == "span 1"
    assert envelopes[0].items[0].payload.json["items"][1]["name"] == "span 2"
    assert len(envelopes[1].items[0].payload.json["items"]) == 2
    assert envelopes[1].items[0].payload.json["items"][0]["name"] == "span 3"
    assert envelopes[1].items[0].payload.json["items"][1]["name"] == "span 4"
    assert len(envelopes[2].items[0].payload.json["items"]) == 1
    assert envelopes[2].items[0].payload.json["items"][0]["name"] == "span 5"


def test_drop_after_max_reached(
    sentry_init, capture_envelopes, capture_record_lost_event_calls, monkeypatch
):
    """New spans are dropped if a bucket reaches MAX_BEFORE_DROP spans."""
    monkeypatch.setattr(SpanBatcher, "MAX_BEFORE_DROP", 2)
    # set the time-based flush limit to something huge so that we're not flushing
    # prematurely
    monkeypatch.setattr(SpanBatcher, "FLUSH_WAIT_TIME", 100000)

    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    envelopes = capture_envelopes()
    record_lost_event_calls = capture_record_lost_event_calls()

    with sentry_sdk.traces.start_span(name="span 1"):
        pass
    with sentry_sdk.traces.start_span(name="span 2"):
        pass
    with sentry_sdk.traces.start_span(name="span 3"):
        pass

    sentry_sdk.flush()

    assert len(envelopes) == 1

    assert len(envelopes[0].items[0].payload.json["items"]) == 2
    assert envelopes[0].items[0].payload.json["items"][0]["name"] == "span 1"
    assert envelopes[0].items[0].payload.json["items"][1]["name"] == "span 2"

    assert ("queue_overflow", "span", None, 1) in record_lost_event_calls


def test_drop_isolated_per_bucket(
    sentry_init, capture_envelopes, capture_record_lost_event_calls, monkeypatch
):
    """A bucket reaching MAX_BEFORE_DROP only drops spans from that trace."""
    monkeypatch.setattr(SpanBatcher, "MAX_BEFORE_DROP", 2)
    monkeypatch.setattr(SpanBatcher, "FLUSH_WAIT_TIME", 100000)

    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    envelopes = capture_envelopes()
    record_lost_event_calls = capture_record_lost_event_calls()

    sentry_sdk.traces.new_trace()
    with sentry_sdk.traces.start_span(name="a1") as span_a:
        trace_id_a = span_a.trace_id
    with sentry_sdk.traces.start_span(name="a2"):
        pass
    with sentry_sdk.traces.start_span(name="a3"):
        pass

    sentry_sdk.traces.new_trace()
    with sentry_sdk.traces.start_span(name="b1") as span_b:
        trace_id_b = span_b.trace_id
    with sentry_sdk.traces.start_span(name="b2"):
        pass

    sentry_sdk.flush()

    assert len(envelopes) == 2

    assert envelopes[0].headers["trace"]["trace_id"] == trace_id_a
    assert len(envelopes[0].items[0].payload.json["items"]) == 2
    assert envelopes[0].items[0].payload.json["items"][0]["name"] == "a1"
    assert envelopes[0].items[0].payload.json["items"][1]["name"] == "a2"

    assert envelopes[1].headers["trace"]["trace_id"] == trace_id_b
    assert len(envelopes[1].items[0].payload.json["items"]) == 2
    assert envelopes[1].items[0].payload.json["items"][0]["name"] == "b1"
    assert envelopes[1].items[0].payload.json["items"][1]["name"] == "b2"

    assert record_lost_event_calls.count(("queue_overflow", "span", None, 1)) == 1


def test_length_based_flushing(sentry_init, capture_items, monkeypatch):
    """A flush event is triggered when a bucket contains MAX_BEFORE_FLUSH spans."""
    monkeypatch.setattr(SpanBatcher, "MAX_BEFORE_FLUSH", 1)
    # set the time-based flush limit to something huge so that we're not hitting
    # it since we want to test MAX_BEFORE_FLUSH instead
    monkeypatch.setattr(SpanBatcher, "FLUSH_WAIT_TIME", 100000)

    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    with sentry_sdk.traces.start_span(name="span"):
        pass

    time.sleep(0.1)

    assert len(items) == 1
    assert items[0].payload["name"] == "span"


def test_weight_based_flushing(sentry_init, capture_envelopes, monkeypatch):
    """When a bucket reaches MAX_BYTES_BEFORE_FLUSH, it'll be flushed."""
    monkeypatch.setattr(SpanBatcher, "MAX_BYTES_BEFORE_FLUSH", 1)
    # set the time-based flush limit to something huge so that it doesn't
    # interfere
    monkeypatch.setattr(SpanBatcher, "FLUSH_WAIT_TIME", 100000)

    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    envelopes = capture_envelopes()

    with sentry_sdk.traces.start_span(name="span"):
        pass

    time.sleep(0.1)

    assert len(envelopes) == 1

    assert len(envelopes[0].items[0].payload.json["items"]) == 1
    assert envelopes[0].items[0].payload.json["items"][0]["name"] == "span"


def test_weight_based_flushing_by_attribute_size(
    sentry_init, capture_envelopes, monkeypatch
):
    """Attribute size is taken into account when using MAX_BYTES_BEFORE_FLUSH."""
    monkeypatch.setattr(SpanBatcher, "FLUSH_WAIT_TIME", 100000)
    monkeypatch.setattr(SpanBatcher, "MAX_BEFORE_FLUSH", 100000)

    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    envelopes = capture_envelopes()

    with sentry_sdk.traces.start_span(name="small span") as bare_span:
        pass

    bare_span_size = SpanBatcher._estimate_size(bare_span._to_json())
    big_attr = "x" * bare_span_size

    monkeypatch.setattr(SpanBatcher, "MAX_BYTES_BEFORE_FLUSH", bare_span_size * 3)

    time.sleep(0.1)

    # The first span alone is well under the byte limit, so no flush yet.
    assert len(envelopes) == 0

    with sentry_sdk.traces.start_span(name="big span", attributes={"big": big_attr}):
        pass

    time.sleep(0.1)

    assert len(envelopes) == 1
    assert envelopes[0].items[0].payload.json["items"][0]["name"] == "small span"
    assert envelopes[0].items[0].payload.json["items"][1]["name"] == "big span"


def test_bucket_recreated_after_flush(sentry_init, capture_envelopes, monkeypatch):
    """Spans for a trace that arrive after that trace's bucket was flushed land in a fresh bucket."""
    monkeypatch.setattr(SpanBatcher, "MAX_BEFORE_FLUSH", 2)
    monkeypatch.setattr(SpanBatcher, "FLUSH_WAIT_TIME", 100000)

    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    envelopes = capture_envelopes()

    sentry_sdk.traces.new_trace()

    with sentry_sdk.traces.start_span(name="span 1") as span1:
        trace_id = span1.trace_id
    with sentry_sdk.traces.start_span(name="span 2"):
        pass

    time.sleep(0.1)

    assert len(envelopes) == 1

    with sentry_sdk.traces.start_span(name="span 3"):
        pass
    with sentry_sdk.traces.start_span(name="span 4"):
        pass

    time.sleep(0.1)

    assert len(envelopes) == 2

    assert envelopes[0].headers["trace"]["trace_id"] == trace_id
    assert len(envelopes[0].items[0].payload.json["items"]) == 2
    assert envelopes[0].items[0].payload.json["items"][0]["name"] == "span 1"
    assert envelopes[0].items[0].payload.json["items"][0]["trace_id"] == trace_id
    assert envelopes[0].items[0].payload.json["items"][1]["name"] == "span 2"
    assert envelopes[0].items[0].payload.json["items"][1]["trace_id"] == trace_id

    assert envelopes[1].headers["trace"]["trace_id"] == trace_id
    assert len(envelopes[1].items[0].payload.json["items"]) == 2
    assert envelopes[1].items[0].payload.json["items"][0]["name"] == "span 3"
    assert envelopes[1].items[0].payload.json["items"][0]["trace_id"] == trace_id
    assert envelopes[1].items[0].payload.json["items"][1]["name"] == "span 4"
    assert envelopes[1].items[0].payload.json["items"][1]["trace_id"] == trace_id


def test_quiet_buckets_flush_eventually(sentry_init, capture_envelopes, monkeypatch):
    """Even if a bucket doesn't trigger any size limits, it'll get flushed eventually."""
    monkeypatch.setattr(SpanBatcher, "FLUSH_WAIT_TIME", 0.1)
    # these are purposefully high so as to never trigger in this test scenario
    monkeypatch.setattr(SpanBatcher, "MAX_BEFORE_FLUSH", 10000000)
    monkeypatch.setattr(SpanBatcher, "MAX_BYTES_BEFORE_FLUSH", 10000000)

    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    envelopes = capture_envelopes()

    with sentry_sdk.traces.start_span(name="span 1"):
        pass

    time.sleep(0.3)

    assert len(envelopes) == 1

    assert len(envelopes[0].items[0].payload.json["items"]) == 1
    assert envelopes[0].items[0].payload.json["items"][0]["name"] == "span 1"


def test_quiet_buckets_flushed_with_busy_neighbors(
    sentry_init, capture_envelopes, monkeypatch
):
    """When there are both busy and quiet buckets, all buckets get flushed eventually."""
    monkeypatch.setattr(SpanBatcher, "FLUSH_WAIT_TIME", 0.1)
    monkeypatch.setattr(SpanBatcher, "MAX_BEFORE_FLUSH", 5)

    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    envelopes = capture_envelopes()

    sentry_sdk.traces.new_trace()

    with sentry_sdk.traces.start_span(name="span 1") as span1:
        trace_id1 = span1.trace_id

    sentry_sdk.traces.new_trace()

    with sentry_sdk.traces.start_span(name="span 2") as span2:
        trace_id2 = span2.trace_id

    for i in range(3, 10):
        with sentry_sdk.traces.start_span(name=f"span {i}"):
            pass

    time.sleep(0.3)

    assert len(envelopes) >= 2

    # we don't care how exactly the spans are distributed over envelopes
    # (since both the time and length based limits are effective, the
    # distribution might not be stable); just check whether the spans are all
    # there and that each envelope only contains spans from one trace
    seen = set()
    for envelope in envelopes:
        assert envelope.headers["trace"]["trace_id"] in (trace_id1, trace_id2)

        if envelope.headers["trace"]["trace_id"] == trace_id1:
            assert len(envelope.items[0].payload.json["items"]) == 1
            assert envelope.items[0].payload.json["items"][0]["name"] == "span 1"
            seen.add(envelope.items[0].payload.json["items"][0]["name"])

        elif envelope.headers["trace"]["trace_id"] == trace_id2:
            for span in envelope.items[0].payload.json["items"]:
                seen.add(span["name"])

    assert seen == {f"span {i}" for i in range(1, 10)}


def test_transport_format(sentry_init, capture_envelopes):
    sentry_init(
        server_name="test-server",
        release="1.0.0",
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    envelopes = capture_envelopes()

    with sentry_sdk.traces.start_span(name="test"):
        ...

    sentry_sdk.get_client().flush()

    assert len(envelopes) == 1
    assert len(envelopes[0].items) == 1
    item = envelopes[0].items[0]

    assert item.type == "span"
    assert item.headers == {
        "type": "span",
        "item_count": 1,
        "content_type": "application/vnd.sentry.items.span.v2+json",
    }
    assert item.payload.json == {
        "version": 2,
        "items": [
            {
                "trace_id": mock.ANY,
                "span_id": mock.ANY,
                "name": "test",
                "status": "ok",
                "is_segment": True,
                "start_timestamp": mock.ANY,
                "end_timestamp": mock.ANY,
                "attributes": mock.ANY,
            }
        ],
    }
    for attribute, value in item.payload.json["items"][0]["attributes"].items():
        assert isinstance(attribute, str)

        assert isinstance(value, dict)
        assert (
            len(value) == 2
        )  # technically, "unit" is also supported, but we don't currently set it anywhere
        assert "value" in value
        assert "type" in value
        assert value["type"] in ("string", "boolean", "integer", "double", "array")


@pytest.mark.skipif(
    sys.platform == "win32"
    or not hasattr(os, "fork")
    or not hasattr(os, "register_at_fork"),
    reason="requires POSIX fork and os.register_at_fork (Python 3.7+)",
)
def test_span_batcher_lock_reset_in_child_after_fork(sentry_init):
    """Regression test for the SpanBatcher fork-deadlock fix.

    If os.fork() runs while another thread holds SpanBatcher._lock, the
    child inherits the lock locked. The holding thread does not exist in
    the child, so the lock can never be released and _ensure_thread
    deadlocks forever. The after-fork hook must replace the lock with a
    fresh one in the child and reset
    _flusher / _flusher_pid / _span_buffer / _running_size / _active /
    _flush_event.
    """
    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    batcher = sentry_sdk.get_client().span_batcher
    assert batcher is not None

    original_lock = batcher._lock
    original_lock.acquire()

    batcher._span_buffer["test-trace-id"].append(object())
    batcher._running_size["test-trace-id"] = 42
    batcher._active.flag = True
    batcher._flush_event.set()
    batcher._running = False

    pid = os.fork()
    if pid == 0:
        replaced = batcher._lock is not original_lock
        unheld = batcher._lock.acquire(blocking=False)

        flusher_reset = batcher._flusher is None and batcher._flusher_pid is None
        span_buffer_reset = len(batcher._span_buffer) == 0
        running_size_reset = len(batcher._running_size) == 0

        active_reset = not getattr(batcher._active, "flag", False)
        event_reset = not batcher._flush_event.is_set()
        running_reset = batcher._running is True

        os._exit(
            0
            if replaced
            and unheld
            and flusher_reset
            and span_buffer_reset
            and running_size_reset
            and active_reset
            and event_reset
            and running_reset
            else 1
        )

    original_lock.release()
    _, status = os.waitpid(pid, 0)
    assert os.WIFEXITED(status) and os.WEXITSTATUS(status) == 0



import pytest

from sentry_sdk import capture_exception
from sentry_sdk.integrations.gnu_backtrace import GnuBacktraceIntegration

LINES = r"""
0. DB::Exception::Exception(DB::Exception::MessageMasked&&, int, bool) @ 0x000000000bfc38a4 in /usr/bin/clickhouse
1. DB::Exception::Exception<String, String>(int, FormatStringHelperImpl<std::type_identity<String>::type, std::type_identity<String>::type>, String&&, String&&) @ 0x00000000075d242c in /usr/bin/clickhouse
2. DB::ActionsMatcher::visit(DB::ASTIdentifier const&, std::shared_ptr<DB::IAST> const&, DB::ActionsMatcher::Data&) @ 0x0000000010b1c648 in /usr/bin/clickhouse
3. DB::ActionsMatcher::visit(DB::ASTFunction const&, std::shared_ptr<DB::IAST> const&, DB::ActionsMatcher::Data&) @ 0x0000000010b1f58c in /usr/bin/clickhouse
4. DB::ActionsMatcher::visit(DB::ASTFunction const&, std::shared_ptr<DB::IAST> const&, DB::ActionsMatcher::Data&) @ 0x0000000010b1f58c in /usr/bin/clickhouse
5. DB::ActionsMatcher::visit(std::shared_ptr<DB::IAST> const&, DB::ActionsMatcher::Data&) @ 0x0000000010b1c394 in /usr/bin/clickhouse
6. DB::InDepthNodeVisitor<DB::ActionsMatcher, true, false, std::shared_ptr<DB::IAST> const>::doVisit(std::shared_ptr<DB::IAST> const&) @ 0x0000000010b154a0 in /usr/bin/clickhouse
7. DB::ExpressionAnalyzer::getRootActions(std::shared_ptr<DB::IAST> const&, bool, std::shared_ptr<DB::ActionsDAG>&, bool) @ 0x0000000010af83b4 in /usr/bin/clickhouse
8. DB::SelectQueryExpressionAnalyzer::appendSelect(DB::ExpressionActionsChain&, bool) @ 0x0000000010aff168 in /usr/bin/clickhouse
9. DB::ExpressionAnalysisResult::ExpressionAnalysisResult(DB::SelectQueryExpressionAnalyzer&, std::shared_ptr<DB::StorageInMemoryMetadata const> const&, bool, bool, bool, std::shared_ptr<DB::FilterDAGInfo> const&, std::shared_ptr<DB::FilterDAGInfo> const&, DB::Block const&) @ 0x0000000010b05b74 in /usr/bin/clickhouse
10. DB::InterpreterSelectQuery::getSampleBlockImpl() @ 0x00000000111559fc in /usr/bin/clickhouse
11. DB::InterpreterSelectQuery::InterpreterSelectQuery(std::shared_ptr<DB::IAST> const&, std::shared_ptr<DB::Context> const&, std::optional<DB::Pipe>, std::shared_ptr<DB::IStorage> const&, DB::SelectQueryOptions const&, std::vector<String, std::allocator<String>> const&, std::shared_ptr<DB::StorageInMemoryMetadata const> const&, std::shared_ptr<DB::PreparedSets>)::$_0::operator()(bool) const @ 0x0000000011148254 in /usr/bin/clickhouse
12. DB::InterpreterSelectQuery::InterpreterSelectQuery(std::shared_ptr<DB::IAST> const&, std::shared_ptr<DB::Context> const&, std::optional<DB::Pipe>, std::shared_ptr<DB::IStorage> const&, DB::SelectQueryOptions const&, std::vector<String, std::allocator<String>> const&, std::shared_ptr<DB::StorageInMemoryMetadata const> const&, std::shared_ptr<DB::PreparedSets>) @ 0x00000000111413e8 in /usr/bin/clickhouse
13. DB::InterpreterSelectWithUnionQuery::InterpreterSelectWithUnionQuery(std::shared_ptr<DB::IAST> const&, std::shared_ptr<DB::Context>, DB::SelectQueryOptions const&, std::vector<String, std::allocator<String>> const&) @ 0x00000000111d3708 in /usr/bin/clickhouse
14. DB::InterpreterFactory::get(std::shared_ptr<DB::IAST>&, std::shared_ptr<DB::Context>, DB::SelectQueryOptions const&) @ 0x0000000011100b64 in /usr/bin/clickhouse
15. DB::executeQueryImpl(char const*, char const*, std::shared_ptr<DB::Context>, bool, DB::QueryProcessingStage::Enum, DB::ReadBuffer*) @ 0x00000000114c3f3c in /usr/bin/clickhouse
16. DB::executeQuery(String const&, std::shared_ptr<DB::Context>, bool, DB::QueryProcessingStage::Enum) @ 0x00000000114c0ec8 in /usr/bin/clickhouse
17. DB::TCPHandler::runImpl() @ 0x00000000121bb5d8 in /usr/bin/clickhouse
18. DB::TCPHandler::run() @ 0x00000000121cb728 in /usr/bin/clickhouse
19. Poco::Net::TCPServerConnection::start() @ 0x00000000146d9404 in /usr/bin/clickhouse
20. Poco::Net::TCPServerDispatcher::run() @ 0x00000000146da900 in /usr/bin/clickhouse
21. Poco::PooledThread::run() @ 0x000000001484da7c in /usr/bin/clickhouse
22. Poco::ThreadImpl::runnableEntry(void*) @ 0x000000001484bc24 in /usr/bin/clickhouse
23. start_thread @ 0x0000000000007624 in /usr/lib/aarch64-linux-gnu/libpthread-2.31.so
24. ? @ 0x00000000000d162c in /usr/lib/aarch64-linux-gnu/libc-2.31.so
"""

LINES_NO_PATH = r"""
0. DB::Exception::Exception(DB::Exception::MessageMasked&&, int, bool) @ 0x000000000bfc38a4
1. DB::Exception::Exception<String, String>(int, FormatStringHelperImpl<std::type_identity<String>::type, std::type_identity<String>::type>, String&&, String&&) @ 0x00000000075d242c
2. DB::ActionsMatcher::visit(DB::ASTIdentifier const&, std::shared_ptr<DB::IAST> const&, DB::ActionsMatcher::Data&) @ 0x0000000010b1c648
3. DB::ActionsMatcher::visit(DB::ASTFunction const&, std::shared_ptr<DB::IAST> const&, DB::ActionsMatcher::Data&) @ 0x0000000010b1f58c
4. DB::ActionsMatcher::visit(DB::ASTFunction const&, std::shared_ptr<DB::IAST> const&, DB::ActionsMatcher::Data&) @ 0x0000000010b1f58c
5. DB::ActionsMatcher::visit(std::shared_ptr<DB::IAST> const&, DB::ActionsMatcher::Data&) @ 0x0000000010b1c394
6. DB::InDepthNodeVisitor<DB::ActionsMatcher, true, false, std::shared_ptr<DB::IAST> const>::doVisit(std::shared_ptr<DB::IAST> const&) @ 0x0000000010b154a0
7. DB::ExpressionAnalyzer::getRootActions(std::shared_ptr<DB::IAST> const&, bool, std::shared_ptr<DB::ActionsDAG>&, bool) @ 0x0000000010af83b4
8. DB::SelectQueryExpressionAnalyzer::appendSelect(DB::ExpressionActionsChain&, bool) @ 0x0000000010aff168
9. DB::ExpressionAnalysisResult::ExpressionAnalysisResult(DB::SelectQueryExpressionAnalyzer&, std::shared_ptr<DB::StorageInMemoryMetadata const> const&, bool, bool, bool, std::shared_ptr<DB::FilterDAGInfo> const&, std::shared_ptr<DB::FilterDAGInfo> const&, DB::Block const&) @ 0x0000000010b05b74
10. DB::InterpreterSelectQuery::getSampleBlockImpl() @ 0x00000000111559fc
11. DB::InterpreterSelectQuery::InterpreterSelectQuery(std::shared_ptr<DB::IAST> const&, std::shared_ptr<DB::Context> const&, std::optional<DB::Pipe>, std::shared_ptr<DB::IStorage> const&, DB::SelectQueryOptions const&, std::vector<String, std::allocator<String>> const&, std::shared_ptr<DB::StorageInMemoryMetadata const> const&, std::shared_ptr<DB::PreparedSets>)::$_0::operator()(bool) const @ 0x0000000011148254
12. DB::InterpreterSelectQuery::InterpreterSelectQuery(std::shared_ptr<DB::IAST> const&, std::shared_ptr<DB::Context> const&, std::optional<DB::Pipe>, std::shared_ptr<DB::IStorage> const&, DB::SelectQueryOptions const&, std::vector<String, std::allocator<String>> const&, std::shared_ptr<DB::StorageInMemoryMetadata const> const&, std::shared_ptr<DB::PreparedSets>) @ 0x00000000111413e8
13. DB::InterpreterSelectWithUnionQuery::InterpreterSelectWithUnionQuery(std::shared_ptr<DB::IAST> const&, std::shared_ptr<DB::Context>, DB::SelectQueryOptions const&, std::vector<String, std::allocator<String>> const&) @ 0x00000000111d3708
14. DB::InterpreterFactory::get(std::shared_ptr<DB::IAST>&, std::shared_ptr<DB::Context>, DB::SelectQueryOptions const&) @ 0x0000000011100b64
15. DB::executeQueryImpl(char const*, char const*, std::shared_ptr<DB::Context>, bool, DB::QueryProcessingStage::Enum, DB::ReadBuffer*) @ 0x00000000114c3f3c
16. DB::executeQuery(String const&, std::shared_ptr<DB::Context>, bool, DB::QueryProcessingStage::Enum) @ 0x00000000114c0ec8
17. DB::TCPHandler::runImpl() @ 0x00000000121bb5d8
18. DB::TCPHandler::run() @ 0x00000000121cb728
19. Poco::Net::TCPServerConnection::start() @ 0x00000000146d9404
20. Poco::Net::TCPServerDispatcher::run() @ 0x00000000146da900
21. Poco::PooledThread::run() @ 0x000000001484da7c
22. Poco::ThreadImpl::runnableEntry(void*) @ 0x000000001484bc24
23. start_thread @ 0x0000000000007624
24. ? @ 0x00000000000d162c
"""


@pytest.mark.parametrize(
    "input", LINES.strip().splitlines() + LINES_NO_PATH.strip().splitlines()
)
def test_basic(sentry_init, capture_events, input):
    sentry_init(integrations=[GnuBacktraceIntegration()])
    events = capture_events()

    try:
        raise ValueError(input)
    except ValueError:
        capture_exception()

    (event,) = events
    (exception,) = event["exception"]["values"]

    assert (
        exception["value"]
        == "<stacktrace parsed and removed by GnuBacktraceIntegration>"
    )
    (frame,) = exception["stacktrace"]["frames"][1:]

    assert frame["function"]



"""
# GCP Cloud Functions unit tests

"""

import json
import os
import os.path
import subprocess
import sys
import tempfile
from textwrap import dedent

import pytest

from sentry_sdk.traces import SpanStatus

FUNCTIONS_PRELUDE = """
from unittest.mock import Mock
import __main__ as gcp_functions
import os

# Initializing all the necessary environment variables
os.environ["FUNCTION_TIMEOUT_SEC"] = "3"
os.environ["FUNCTION_NAME"] = "Google Cloud function"
os.environ["ENTRY_POINT"] = "cloud_function"
os.environ["FUNCTION_IDENTITY"] = "func_ID"
os.environ["FUNCTION_REGION"] = "us-central1"
os.environ["GCP_PROJECT"] = "serverless_project"

def log_return_value(func):
    def inner(*args, **kwargs):
        rv = func(*args, **kwargs)

        print("\\nRETURN VALUE: {}\\n".format(json.dumps(rv)))

        return rv

    return inner

gcp_functions.worker_v1 = Mock()
gcp_functions.worker_v1.FunctionHandler = Mock()
gcp_functions.worker_v1.FunctionHandler.invoke_user_function = log_return_value(cloud_function)


import sentry_sdk
from sentry_sdk.integrations.gcp import GcpIntegration
import json
import time

from sentry_sdk.transport import HttpTransport

def event_processor(event):
    # Adding delay which would allow us to capture events.
    time.sleep(1)
    return event

class TestTransport(HttpTransport):
    def capture_envelope(self, envelope):
        for item in envelope.items:
            if item.headers.get("type") == "span":
                payload = item.payload.json
                for span_item in payload.get("items", []):
                    attrs = {k: v["value"] for k, v in span_item.get("attributes", {}).items()}
                    span_data = {k: v for k, v in span_item.items() if k != "attributes"}
                    span_data["attributes"] = attrs
                    print("\\nSPAN: {}\\n".format(json.dumps(span_data)))
            else:
                envelope_bytes = item.get_bytes()
                print("\\nENVELOPE: {}\\n".format(envelope_bytes.decode(\"utf-8\")))


def init_sdk(timeout_warning=False, **extra_init_args):
    sentry_sdk.init(
        dsn="https://123abc@example.com/123",
        transport=TestTransport,
        integrations=[GcpIntegration(timeout_warning=timeout_warning)],
        shutdown_timeout=10,
        # excepthook -> dedupe -> event_processor client report gets added
        # which we don't really care about for these tests
        send_client_reports=False,
        **extra_init_args
    )

"""


@pytest.fixture
def run_cloud_function():
    def inner(code, subprocess_kwargs=()):
        envelope_items = []
        span_items = []
        return_value = None

        # STEP : Create a zip of cloud function

        subprocess_kwargs = dict(subprocess_kwargs)

        with tempfile.TemporaryDirectory() as tmpdir:
            main_py = os.path.join(tmpdir, "main.py")
            with open(main_py, "w") as f:
                f.write(code)

            setup_cfg = os.path.join(tmpdir, "setup.cfg")

            with open(setup_cfg, "w") as f:
                f.write("[install]\nprefix=")

            subprocess.check_call(
                [
                    "uv",
                    "build",
                    "--sdist",
                    "-p",
                    sys.executable,
                    "-o",
                    os.path.join(tmpdir, ".."),
                ],
                **subprocess_kwargs,
            )

            subprocess.check_call(
                "uv pip install -p {} ../*.tar.gz --target .".format(sys.executable),
                cwd=tmpdir,
                shell=True,
                **subprocess_kwargs,
            )

            stream = os.popen("python {}/main.py".format(tmpdir))
            stream_data = stream.read()

            stream.close()

            for line in stream_data.splitlines():
                print("GCP:", line)
                if line.startswith("ENVELOPE: "):
                    line = line[len("ENVELOPE: ") :]
                    envelope_items.append(json.loads(line))
                elif line.startswith("SPAN: "):
                    line = line[len("SPAN: ") :]
                    span_items.append(json.loads(line))
                elif line.startswith("RETURN VALUE: "):
                    line = line[len("RETURN VALUE: ") :]
                    return_value = json.loads(line)
                else:
                    continue

            stream.close()

        return envelope_items, return_value, span_items

    return inner


def test_handled_exception(run_cloud_function):
    envelope_items, return_value, _ = run_cloud_function(
        dedent(
            """
        functionhandler = None
        event = {}
        def cloud_function(functionhandler, event):
            raise Exception("something went wrong")
        """
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(timeout_warning=False)
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )
    assert envelope_items[0]["level"] == "error"
    (exception,) = envelope_items[0]["exception"]["values"]

    assert exception["type"] == "Exception"
    assert exception["value"] == "something went wrong"
    assert exception["mechanism"]["type"] == "gcp"
    assert not exception["mechanism"]["handled"]


def test_unhandled_exception(run_cloud_function):
    envelope_items, _, _ = run_cloud_function(
        dedent(
            """
        functionhandler = None
        event = {}
        def cloud_function(functionhandler, event):
            x = 3/0
            return "3"
        """
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(timeout_warning=False)
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )
    assert envelope_items[0]["level"] == "error"
    (exception,) = envelope_items[0]["exception"]["values"]

    assert exception["type"] == "ZeroDivisionError"
    assert exception["value"] == "division by zero"
    assert exception["mechanism"]["type"] == "gcp"
    assert not exception["mechanism"]["handled"]


def test_timeout_error(run_cloud_function):
    envelope_items, _, _ = run_cloud_function(
        dedent(
            """
        functionhandler = None
        event = {}
        def cloud_function(functionhandler, event):
            sentry_sdk.set_tag("cloud_function", "true")
            time.sleep(10)
            return "3"
        """
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(timeout_warning=True)
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )
    assert envelope_items[0]["level"] == "error"
    (exception,) = envelope_items[0]["exception"]["values"]

    assert exception["type"] == "ServerlessTimeoutWarning"
    assert (
        exception["value"]
        == "WARNING : Function is expected to get timed out. Configured timeout duration = 3 seconds."
    )
    assert exception["mechanism"]["type"] == "threading"
    assert not exception["mechanism"]["handled"]

    assert envelope_items[0]["tags"]["cloud_function"] == "true"


def test_performance_no_error(run_cloud_function):
    envelope_items, _, _ = run_cloud_function(
        dedent(
            """
        functionhandler = None
        event = {}
        def cloud_function(functionhandler, event):
            return "test_string"
        """
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(traces_sample_rate=1.0)
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )

    assert envelope_items[0]["type"] == "transaction"
    assert envelope_items[0]["contexts"]["trace"]["op"] == "function.gcp"
    assert envelope_items[0]["transaction"].startswith("Google Cloud function")
    assert envelope_items[0]["transaction_info"] == {"source": "component"}
    assert envelope_items[0]["transaction"] in envelope_items[0]["request"]["url"]


def test_performance_error(run_cloud_function):
    envelope_items, _, _ = run_cloud_function(
        dedent(
            """
        functionhandler = None
        event = {}
        def cloud_function(functionhandler, event):
            raise Exception("something went wrong")
        """
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(traces_sample_rate=1.0)
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )

    assert envelope_items[0]["level"] == "error"
    (exception,) = envelope_items[0]["exception"]["values"]

    assert exception["type"] == "Exception"
    assert exception["value"] == "something went wrong"
    assert exception["mechanism"]["type"] == "gcp"
    assert not exception["mechanism"]["handled"]

    assert envelope_items[1]["type"] == "transaction"
    assert envelope_items[1]["contexts"]["trace"]["op"] == "function.gcp"
    assert envelope_items[1]["transaction"].startswith("Google Cloud function")
    assert envelope_items[1]["transaction"] in envelope_items[0]["request"]["url"]


def test_traces_sampler_gets_correct_values_in_sampling_context(
    run_cloud_function,
    DictionaryContaining,  # noqa:N803
):
    # TODO: There are some decent sized hacks below. For more context, see the
    # long comment in the test of the same name in the AWS integration. The
    # situations there and here aren't identical, but they're similar enough
    # that solving one would probably solve both.

    import inspect

    _, return_value, _ = run_cloud_function(
        dedent(
            """
            functionhandler = None
            event = {
                "type": "chase",
                "chasers": ["Maisey", "Charlie"],
                "num_squirrels": 2,
            }
            def cloud_function(functionhandler, event):
                # this runs after the transaction has started, which means we
                # can make assertions about traces_sampler
                try:
                    traces_sampler.assert_any_call(
                        DictionaryContaining({
                            "gcp_env": DictionaryContaining({
                                "function_name": "chase_into_tree",
                                "function_region": "dogpark",
                                "function_project": "SquirrelChasing",
                            }),
                            "gcp_event": {
                                "type": "chase",
                                "chasers": ["Maisey", "Charlie"],
                                "num_squirrels": 2,
                            },
                        })
                    )
                except AssertionError:
                    # catch the error and return it because the error itself will
                    # get swallowed by the SDK as an "internal exception"
                    return {"AssertionError raised": True,}

                return {"AssertionError raised": False,}
            """
        )
        + FUNCTIONS_PRELUDE
        + dedent(inspect.getsource(DictionaryContaining))
        + dedent(
            """
            os.environ["FUNCTION_NAME"] = "chase_into_tree"
            os.environ["FUNCTION_REGION"] = "dogpark"
            os.environ["GCP_PROJECT"] = "SquirrelChasing"

            def _safe_is_equal(x, y):
                # copied from conftest.py - see docstring and comments there
                try:
                    is_equal = x.__eq__(y)
                except AttributeError:
                    is_equal = NotImplemented

                if is_equal == NotImplemented:
                    return x == y

                return is_equal

            traces_sampler = Mock(return_value=True)

            init_sdk(
                traces_sampler=traces_sampler,
            )

            gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
            """
        )
    )

    assert return_value["AssertionError raised"] is False


def test_error_has_new_trace_context_performance_enabled(run_cloud_function):
    """
    Check if an 'trace' context is added to errros and transactions when performance monitoring is enabled.
    """
    envelope_items, _, _ = run_cloud_function(
        dedent(
            """
        functionhandler = None
        event = {}
        def cloud_function(functionhandler, event):
            sentry_sdk.capture_message("hi")
            x = 3/0
            return "3"
        """
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(traces_sample_rate=1.0)
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )
    (msg_event, error_event, transaction_event) = envelope_items

    assert "trace" in msg_event["contexts"]
    assert "trace_id" in msg_event["contexts"]["trace"]

    assert "trace" in error_event["contexts"]
    assert "trace_id" in error_event["contexts"]["trace"]

    assert "trace" in transaction_event["contexts"]
    assert "trace_id" in transaction_event["contexts"]["trace"]

    assert (
        msg_event["contexts"]["trace"]["trace_id"]
        == error_event["contexts"]["trace"]["trace_id"]
        == transaction_event["contexts"]["trace"]["trace_id"]
    )


def test_error_has_new_trace_context_performance_disabled(run_cloud_function):
    """
    Check if an 'trace' context is added to errros and transactions when performance monitoring is disabled.
    """
    envelope_items, _, _ = run_cloud_function(
        dedent(
            """
        functionhandler = None
        event = {}
        def cloud_function(functionhandler, event):
            sentry_sdk.capture_message("hi")
            x = 3/0
            return "3"
        """
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(traces_sample_rate=None),  # this is the default, just added for clarity
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )

    (msg_event, error_event) = envelope_items

    assert "trace" in msg_event["contexts"]
    assert "trace_id" in msg_event["contexts"]["trace"]

    assert "trace" in error_event["contexts"]
    assert "trace_id" in error_event["contexts"]["trace"]

    assert (
        msg_event["contexts"]["trace"]["trace_id"]
        == error_event["contexts"]["trace"]["trace_id"]
    )


def test_error_has_existing_trace_context_performance_enabled(run_cloud_function):
    """
    Check if an 'trace' context is added to errros and transactions
    from the incoming 'sentry-trace' header when performance monitoring is enabled.
    """
    trace_id = "471a43a4192642f0b136d5159a501701"
    parent_span_id = "6e8f22c393e68f19"
    parent_sampled = 1
    sentry_trace_header = "{}-{}-{}".format(trace_id, parent_span_id, parent_sampled)

    envelope_items, _, _ = run_cloud_function(
        dedent(
            """
        functionhandler = None

        from collections import namedtuple
        GCPEvent = namedtuple("GCPEvent", ["headers"])
        event = GCPEvent(headers={"sentry-trace": "%s"})

        def cloud_function(functionhandler, event):
            sentry_sdk.capture_message("hi")
            x = 3/0
            return "3"
        """
            % sentry_trace_header
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(traces_sample_rate=1.0)
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )
    (msg_event, error_event, transaction_event) = envelope_items

    assert "trace" in msg_event["contexts"]
    assert "trace_id" in msg_event["contexts"]["trace"]

    assert "trace" in error_event["contexts"]
    assert "trace_id" in error_event["contexts"]["trace"]

    assert "trace" in transaction_event["contexts"]
    assert "trace_id" in transaction_event["contexts"]["trace"]

    assert (
        msg_event["contexts"]["trace"]["trace_id"]
        == error_event["contexts"]["trace"]["trace_id"]
        == transaction_event["contexts"]["trace"]["trace_id"]
        == "471a43a4192642f0b136d5159a501701"
    )


def test_error_has_existing_trace_context_performance_disabled(run_cloud_function):
    """
    Check if an 'trace' context is added to errros and transactions
    from the incoming 'sentry-trace' header when performance monitoring is disabled.
    """
    trace_id = "471a43a4192642f0b136d5159a501701"
    parent_span_id = "6e8f22c393e68f19"
    parent_sampled = 1
    sentry_trace_header = "{}-{}-{}".format(trace_id, parent_span_id, parent_sampled)

    envelope_items, _, _ = run_cloud_function(
        dedent(
            """
        functionhandler = None

        from collections import namedtuple
        GCPEvent = namedtuple("GCPEvent", ["headers"])
        event = GCPEvent(headers={"sentry-trace": "%s"})

        def cloud_function(functionhandler, event):
            sentry_sdk.capture_message("hi")
            x = 3/0
            return "3"
        """
            % sentry_trace_header
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(traces_sample_rate=None),  # this is the default, just added for clarity
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )
    (msg_event, error_event) = envelope_items

    assert "trace" in msg_event["contexts"]
    assert "trace_id" in msg_event["contexts"]["trace"]

    assert "trace" in error_event["contexts"]
    assert "trace_id" in error_event["contexts"]["trace"]

    assert (
        msg_event["contexts"]["trace"]["trace_id"]
        == error_event["contexts"]["trace"]["trace_id"]
        == "471a43a4192642f0b136d5159a501701"
    )


def test_span_origin(run_cloud_function):
    events, _, _ = run_cloud_function(
        dedent(
            """
        functionhandler = None
        event = {}
        def cloud_function(functionhandler, event):
            return "test_string"
        """
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(traces_sample_rate=1.0)
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )

    (event,) = events

    assert event["contexts"]["trace"]["origin"] == "auto.function.gcp"


def test_span_streaming_no_error(run_cloud_function):
    _, _, span_items = run_cloud_function(
        dedent(
            """
        functionhandler = None
        event = {}
        def cloud_function(functionhandler, event):
            return "test_string"
        """
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(traces_sample_rate=1.0, trace_lifecycle="stream")
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )

    assert len(span_items) == 1
    segment_span = span_items[0]

    assert segment_span["is_segment"] is True
    assert segment_span["name"] == "Google Cloud function"
    assert segment_span["attributes"]["sentry.op"] == "function.gcp"
    assert segment_span["attributes"]["sentry.origin"] == "auto.function.gcp"
    assert segment_span["attributes"]["sentry.segment.name.source"] == "component"
    assert segment_span["attributes"]["cloud.provider"] == "gcp"
    assert segment_span["attributes"]["faas.name"] == "Google Cloud function"
    assert segment_span["attributes"]["gcp.project.id"] == "serverless_project"
    assert segment_span["attributes"]["faas.identity"] == "func_ID"
    assert segment_span["attributes"]["faas.entry_point"] == "cloud_function"


def test_span_streaming_error(run_cloud_function):
    envelope_items, _, span_items = run_cloud_function(
        dedent(
            """
        functionhandler = None
        event = {}
        def cloud_function(functionhandler, event):
            raise Exception("something went wrong")
        """
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(traces_sample_rate=1.0, trace_lifecycle="stream")
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )

    assert envelope_items[0]["level"] == "error"
    (exception,) = envelope_items[0]["exception"]["values"]
    assert exception["type"] == "Exception"
    assert exception["value"] == "something went wrong"
    assert exception["mechanism"]["type"] == "gcp"
    assert not exception["mechanism"]["handled"]

    assert len(span_items) == 1
    segment_span = span_items[0]
    assert segment_span["is_segment"] is True
    assert segment_span["name"] == "Google Cloud function"
    assert segment_span["attributes"]["sentry.op"] == "function.gcp"
    assert segment_span["attributes"]["sentry.origin"] == "auto.function.gcp"
    assert segment_span["attributes"]["sentry.segment.name.source"] == "component"
    assert segment_span["attributes"]["cloud.provider"] == "gcp"
    assert segment_span["attributes"]["faas.name"] == "Google Cloud function"
    assert segment_span["attributes"]["gcp.project.id"] == "serverless_project"
    assert segment_span["attributes"]["faas.identity"] == "func_ID"
    assert segment_span["attributes"]["faas.entry_point"] == "cloud_function"
    assert segment_span["status"] == SpanStatus.ERROR


def test_span_streaming_existing_trace_context(run_cloud_function):
    trace_id = "471a43a4192642f0b136d5159a501701"
    parent_span_id = "6e8f22c393e68f19"
    parent_sampled = 1
    sentry_trace_header = "{}-{}-{}".format(trace_id, parent_span_id, parent_sampled)

    envelope_items, _, span_items = run_cloud_function(
        dedent(
            """
        functionhandler = None

        from collections import namedtuple
        GCPEvent = namedtuple("GCPEvent", ["headers"])
        event = GCPEvent(headers={"sentry-trace": "%s"})

        def cloud_function(functionhandler, event):
            sentry_sdk.capture_message("hi")
            return "ok"
        """
            % sentry_trace_header
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(traces_sample_rate=1.0, trace_lifecycle="stream")
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )

    (msg_event,) = envelope_items
    assert msg_event["contexts"]["trace"]["trace_id"] == trace_id

    assert len(span_items) == 1
    segment_span = span_items[0]
    assert segment_span["is_segment"] is True
    assert segment_span["trace_id"] == trace_id
    assert segment_span["name"] == "Google Cloud function"
    assert segment_span["attributes"]["sentry.op"] == "function.gcp"
    assert segment_span["attributes"]["sentry.origin"] == "auto.function.gcp"
    assert segment_span["attributes"]["sentry.segment.name.source"] == "component"
    assert segment_span["attributes"]["cloud.provider"] == "gcp"
    assert segment_span["attributes"]["faas.name"] == "Google Cloud function"
    assert segment_span["attributes"]["gcp.project.id"] == "serverless_project"
    assert segment_span["attributes"]["faas.identity"] == "func_ID"
    assert segment_span["attributes"]["faas.entry_point"] == "cloud_function"


def test_span_streaming_request_attributes(run_cloud_function):
    _, _, span_items = run_cloud_function(
        dedent(
            """
        functionhandler = None

        from collections import namedtuple
        GCPEvent = namedtuple("GCPEvent", ["headers", "method", "query_string"])
        event = GCPEvent(
            headers={"content-type": "application/json", "accept": "text/html"},
            method="POST",
            query_string=b"foo=bar",
        )

        def cloud_function(functionhandler, event):
            return "ok"
        """
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(traces_sample_rate=1.0, send_default_pii=True, trace_lifecycle="stream")
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )

    assert len(span_items) == 1
    segment_span = span_items[0]
    attrs = segment_span["attributes"]

    assert attrs["http.request.method"] == "POST"
    assert attrs["url.query"] == "foo=bar"
    assert attrs["http.request.header.content-type"] == "application/json"
    assert attrs["http.request.header.accept"] == "text/html"
    assert attrs["faas.name"] == "Google Cloud function"
    assert attrs["gcp.project.id"] == "serverless_project"
    assert attrs["faas.identity"] == "func_ID"
    assert attrs["faas.entry_point"] == "cloud_function"


def test_span_streaming_no_query_string_without_pii(run_cloud_function):
    _, _, span_items = run_cloud_function(
        dedent(
            """
        functionhandler = None

        from collections import namedtuple
        GCPEvent = namedtuple("GCPEvent", ["headers", "method", "query_string"])
        event = GCPEvent(
            headers={},
            method="GET",
            query_string=b"secret=hunter2",
        )

        def cloud_function(functionhandler, event):
            return "ok"
        """
        )
        + FUNCTIONS_PRELUDE
        + dedent(
            """
        init_sdk(traces_sample_rate=1.0, send_default_pii=False, trace_lifecycle="stream")
        gcp_functions.worker_v1.FunctionHandler.invoke_user_function(functionhandler, event)
        """
        )
    )

    assert len(span_items) == 1
    segment_span = span_items[0]
    attrs = segment_span["attributes"]

    assert "url.query" not in attrs
    assert attrs["http.request.method"] == "GET"
    assert attrs["faas.name"] == "Google Cloud function"
    assert attrs["gcp.project.id"] == "serverless_project"
    assert attrs["faas.identity"] == "func_ID"
    assert attrs["faas.entry_point"] == "cloud_function"



from typing import AsyncGenerator, Optional

import pytest

import sentry_sdk

strawberry = pytest.importorskip("strawberry")
pytest.importorskip("fastapi")
pytest.importorskip("flask")


from fastapi import FastAPI
from fastapi.testclient import TestClient
from flask import Flask
from strawberry.fastapi import GraphQLRouter
from strawberry.flask.views import GraphQLView

from sentry_sdk.consts import OP
from sentry_sdk.integrations.fastapi import FastApiIntegration
from sentry_sdk.integrations.flask import FlaskIntegration
from sentry_sdk.integrations.starlette import StarletteIntegration
from sentry_sdk.integrations.strawberry import (
    SentryAsyncExtension,
    SentrySyncExtension,
    StrawberryIntegration,
)
from tests.conftest import ApproxDict

try:
    from strawberry.extensions.tracing import (
        SentryTracingExtension,
        SentryTracingExtensionSync,
    )
except ImportError:
    SentryTracingExtension = None
    SentryTracingExtensionSync = None

parameterize_strawberry_test = pytest.mark.parametrize(
    "client_factory,async_execution,framework_integrations",
    (
        (
            "async_app_client_factory",
            True,
            [FastApiIntegration(), StarletteIntegration()],
        ),
        ("sync_app_client_factory", False, [FlaskIntegration()]),
    ),
)


@strawberry.type
class Query:
    @strawberry.field
    def hello(self) -> str:
        return "Hello World"

    @strawberry.field
    def error(self) -> int:
        return 1 / 0


@strawberry.type
class Mutation:
    @strawberry.mutation
    def change(self, attribute: str) -> str:
        return attribute


@strawberry.type
class Message:
    content: str


@strawberry.type
class Subscription:
    @strawberry.subscription
    async def message_added(self) -> Optional[AsyncGenerator[Message, None]]:
        message = Message(content="Hello, world!")
        yield message


@pytest.fixture
def async_app_client_factory():
    def create_app(schema):
        async_app = FastAPI()
        async_app.include_router(GraphQLRouter(schema), prefix="/graphql")
        return TestClient(async_app)

    return create_app


@pytest.fixture
def sync_app_client_factory():
    def create_app(schema):
        sync_app = Flask(__name__)
        sync_app.add_url_rule(
            "/graphql",
            view_func=GraphQLView.as_view("graphql_view", schema=schema),
        )
        return sync_app.test_client()

    return create_app


def test_async_execution_uses_async_extension(sentry_init):
    sentry_init(integrations=[StrawberryIntegration(async_execution=True)])

    schema = strawberry.Schema(Query)
    assert SentryAsyncExtension in schema.extensions
    assert SentrySyncExtension not in schema.extensions


def test_sync_execution_uses_sync_extension(sentry_init):
    sentry_init(integrations=[StrawberryIntegration(async_execution=False)])

    schema = strawberry.Schema(Query)
    assert SentrySyncExtension in schema.extensions
    assert SentryAsyncExtension not in schema.extensions


def test_use_sync_extension_if_not_specified(sentry_init):
    sentry_init(integrations=[StrawberryIntegration()])
    schema = strawberry.Schema(Query)
    assert SentrySyncExtension in schema.extensions
    assert SentryAsyncExtension not in schema.extensions


@pytest.mark.skipif(
    SentryTracingExtension is None,
    reason="SentryTracingExtension no longer available in this Strawberry version",
)
def test_replace_existing_sentry_async_extension(sentry_init):
    sentry_init(integrations=[StrawberryIntegration()])

    schema = strawberry.Schema(Query, extensions=[SentryTracingExtension])
    assert SentryTracingExtension not in schema.extensions
    assert SentrySyncExtension not in schema.extensions
    assert SentryAsyncExtension in schema.extensions


@pytest.mark.skipif(
    SentryTracingExtensionSync is None,
    reason="SentryTracingExtensionSync no longer available in this Strawberry version",
)
def test_replace_existing_sentry_sync_extension(sentry_init):
    sentry_init(integrations=[StrawberryIntegration()])

    schema = strawberry.Schema(Query, extensions=[SentryTracingExtensionSync])
    assert SentryTracingExtensionSync not in schema.extensions
    assert SentryAsyncExtension not in schema.extensions
    assert SentrySyncExtension in schema.extensions


@parameterize_strawberry_test
def test_capture_request_if_available_and_send_pii_is_on(
    request,
    sentry_init,
    capture_events,
    client_factory,
    async_execution,
    framework_integrations,
):
    sentry_init(
        send_default_pii=True,
        integrations=[
            StrawberryIntegration(async_execution=async_execution),
        ]
        + framework_integrations,
    )
    events = capture_events()

    schema = strawberry.Schema(Query)

    client_factory = request.getfixturevalue(client_factory)
    client = client_factory(schema)

    query = "query ErrorQuery { error }"
    client.post("/graphql", json={"query": query, "operationName": "ErrorQuery"})

    assert len(events) == 1

    (error_event,) = events

    assert error_event["exception"]["values"][0]["mechanism"]["type"] == "strawberry"
    assert error_event["request"]["api_target"] == "graphql"
    assert error_event["request"]["data"] == {
        "query": query,
        "operationName": "ErrorQuery",
    }
    assert error_event["contexts"]["response"] == {
        "data": {
            "data": None,
            "errors": [
                {
                    "message": "division by zero",
                    "locations": [{"line": 1, "column": 20}],
                    "path": ["error"],
                }
            ],
        }
    }
    assert len(error_event["breadcrumbs"]["values"]) == 1
    assert error_event["breadcrumbs"]["values"][0]["category"] == "graphql.operation"
    assert error_event["breadcrumbs"]["values"][0]["data"] == {
        "operation_name": "ErrorQuery",
        "operation_type": "query",
    }


@parameterize_strawberry_test
def test_do_not_capture_request_if_send_pii_is_off(
    request,
    sentry_init,
    capture_events,
    client_factory,
    async_execution,
    framework_integrations,
):
    sentry_init(
        integrations=[
            StrawberryIntegration(async_execution=async_execution),
        ]
        + framework_integrations,
    )
    events = capture_events()

    schema = strawberry.Schema(Query)

    client_factory = request.getfixturevalue(client_factory)
    client = client_factory(schema)

    query = "query ErrorQuery { error }"
    client.post("/graphql", json={"query": query, "operationName": "ErrorQuery"})

    assert len(events) == 1

    (error_event,) = events
    assert error_event["exception"]["values"][0]["mechanism"]["type"] == "strawberry"
    assert "data" not in error_event["request"]
    assert "response" not in error_event["contexts"]

    assert len(error_event["breadcrumbs"]["values"]) == 1
    assert error_event["breadcrumbs"]["values"][0]["category"] == "graphql.operation"
    assert error_event["breadcrumbs"]["values"][0]["data"] == {
        "operation_name": "ErrorQuery",
        "operation_type": "query",
    }


@parameterize_strawberry_test
def test_breadcrumb_no_operation_name(
    request,
    sentry_init,
    capture_events,
    client_factory,
    async_execution,
    framework_integrations,
):
    sentry_init(
        integrations=[
            StrawberryIntegration(async_execution=async_execution),
        ]
        + framework_integrations,
    )
    events = capture_events()

    schema = strawberry.Schema(Query)

    client_factory = request.getfixturevalue(client_factory)
    client = client_factory(schema)

    query = "{ error }"
    client.post("/graphql", json={"query": query})

    assert len(events) == 1

    (error_event,) = events

    assert len(error_event["breadcrumbs"]["values"]) == 1
    assert error_event["breadcrumbs"]["values"][0]["category"] == "graphql.operation"
    assert error_event["breadcrumbs"]["values"][0]["data"] == {
        "operation_name": None,
        "operation_type": "query",
    }


@parameterize_strawberry_test
@pytest.mark.parametrize(
    "send_default_pii",
    [True, False],
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_capture_transaction_on_error(
    request,
    sentry_init,
    capture_events,
    capture_items,
    client_factory,
    async_execution,
    framework_integrations,
    send_default_pii,
    span_streaming,
):
    sentry_init(
        send_default_pii=send_default_pii,
        integrations=[
            StrawberryIntegration(async_execution=async_execution),
        ]
        + framework_integrations,
        traces_sample_rate=1,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("event", "span")
    else:
        events = capture_events()

    schema = strawberry.Schema(Query)

    client_factory = request.getfixturevalue(client_factory)
    client = client_factory(schema)

    query = "query ErrorQuery { error }"
    client.post("/graphql", json={"query": query, "operationName": "ErrorQuery"})

    if span_streaming:
        sentry_sdk.flush()
        error_events = [i.payload for i in items if i.type == "event"]
        spans = [i.payload for i in items if i.type == "span"]

        assert len(error_events) == 1

        assert len(spans) == 5
        parse_span, validate_span, resolve_span, query_span, segment = spans

        assert segment["is_segment"] is True
        assert segment["name"] == "ErrorQuery"
        assert segment["attributes"]["sentry.op"] == OP.GRAPHQL_QUERY

        assert query_span["attributes"]["sentry.op"] == OP.GRAPHQL_QUERY
        assert query_span["name"] == "query ErrorQuery"
        assert query_span["attributes"]["graphql.operation.type"] == "query"
        assert query_span["attributes"]["graphql.operation.name"] == "ErrorQuery"

        if send_default_pii is True:
            assert query_span["attributes"]["graphql.document"] == query
        else:
            assert "graphql.document" not in query_span["attributes"]

        assert parse_span["attributes"]["sentry.op"] == OP.GRAPHQL_PARSE
        assert parse_span["name"] == "parsing"
        assert parse_span["parent_span_id"] == query_span["span_id"]

        assert validate_span["attributes"]["sentry.op"] == OP.GRAPHQL_VALIDATE
        assert validate_span["name"] == "validation"
        assert validate_span["parent_span_id"] == query_span["span_id"]

        assert resolve_span["attributes"]["sentry.op"] == OP.GRAPHQL_RESOLVE
        assert resolve_span["name"] == "resolving Query.error"
        assert resolve_span["parent_span_id"] == query_span["span_id"]
    else:
        assert len(events) == 2
        (_, transaction_event) = events

        assert transaction_event["transaction"] == "ErrorQuery"
        assert transaction_event["contexts"]["trace"]["op"] == OP.GRAPHQL_QUERY
        assert transaction_event["spans"]

        query_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_QUERY
        ]
        assert len(query_spans) == 1, "exactly one query span expected"
        query_span = query_spans[0]
        assert query_span["description"] == "query ErrorQuery"
        assert query_span["data"]["graphql.operation.type"] == "query"
        assert query_span["data"]["graphql.operation.name"] == "ErrorQuery"
        assert query_span["data"]["graphql.resource_name"]

        if send_default_pii is True:
            assert query_span["data"]["graphql.document"] == query
        else:
            assert "graphql.document" not in query_span["data"]

        parse_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_PARSE
        ]
        assert len(parse_spans) == 1, "exactly one parse span expected"
        parse_span = parse_spans[0]
        assert parse_span["parent_span_id"] == query_span["span_id"]
        assert parse_span["description"] == "parsing"

        validate_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_VALIDATE
        ]
        assert len(validate_spans) == 1, "exactly one validate span expected"
        validate_span = validate_spans[0]
        assert validate_span["parent_span_id"] == query_span["span_id"]
        assert validate_span["description"] == "validation"

        resolve_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_RESOLVE
        ]
        assert len(resolve_spans) == 1, "exactly one resolve span expected"
        resolve_span = resolve_spans[0]
        assert resolve_span["parent_span_id"] == query_span["span_id"]
        assert resolve_span["description"] == "resolving Query.error"
        assert resolve_span["data"] == ApproxDict(
            {
                "graphql.field_name": "error",
                "graphql.parent_type": "Query",
                "graphql.field_path": "Query.error",
                "graphql.path": "error",
            }
        )


@parameterize_strawberry_test
@pytest.mark.parametrize(
    "send_default_pii",
    [True, False],
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_capture_transaction_on_success(
    request,
    sentry_init,
    capture_events,
    capture_items,
    client_factory,
    async_execution,
    framework_integrations,
    send_default_pii,
    span_streaming,
):
    sentry_init(
        integrations=[
            StrawberryIntegration(async_execution=async_execution),
        ]
        + framework_integrations,
        traces_sample_rate=1,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    schema = strawberry.Schema(Query)

    client_factory = request.getfixturevalue(client_factory)
    client = client_factory(schema)

    query = "query GreetingQuery { hello }"
    client.post("/graphql", json={"query": query, "operationName": "GreetingQuery"})

    if span_streaming:
        sentry_sdk.flush()
        spans = [i.payload for i in items]

        assert len(spans) == 5
        parse_span, validate_span, resolve_span, query_span, segment = spans

        assert segment["is_segment"] is True
        assert segment["name"] == "GreetingQuery"
        assert segment["attributes"]["sentry.op"] == OP.GRAPHQL_QUERY

        assert query_span["attributes"]["sentry.op"] == OP.GRAPHQL_QUERY
        assert query_span["name"] == "query GreetingQuery"
        assert query_span["attributes"]["graphql.operation.type"] == "query"
        assert query_span["attributes"]["graphql.operation.name"] == "GreetingQuery"

        if send_default_pii is True:
            assert query_span["attributes"]["graphql.document"] == query
        else:
            assert "graphql.document" not in query_span["attributes"]

        assert parse_span["attributes"]["sentry.op"] == OP.GRAPHQL_PARSE
        assert parse_span["name"] == "parsing"
        assert parse_span["parent_span_id"] == query_span["span_id"]

        assert validate_span["attributes"]["sentry.op"] == OP.GRAPHQL_VALIDATE
        assert validate_span["name"] == "validation"
        assert validate_span["parent_span_id"] == query_span["span_id"]

        assert resolve_span["attributes"]["sentry.op"] == OP.GRAPHQL_RESOLVE
        assert resolve_span["name"] == "resolving Query.hello"
        assert resolve_span["parent_span_id"] == query_span["span_id"]
    else:
        assert len(events) == 1
        (transaction_event,) = events

        assert transaction_event["transaction"] == "GreetingQuery"
        assert transaction_event["contexts"]["trace"]["op"] == OP.GRAPHQL_QUERY
        assert transaction_event["spans"]

        query_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_QUERY
        ]
        assert len(query_spans) == 1, "exactly one query span expected"
        query_span = query_spans[0]
        assert query_span["description"] == "query GreetingQuery"
        assert query_span["data"]["graphql.operation.type"] == "query"
        assert query_span["data"]["graphql.operation.name"] == "GreetingQuery"
        assert query_span["data"]["graphql.resource_name"]

        if send_default_pii is True:
            assert query_span["data"]["graphql.document"] == query
        else:
            assert "graphql.document" not in query_span["data"]

        parse_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_PARSE
        ]
        assert len(parse_spans) == 1, "exactly one parse span expected"
        parse_span = parse_spans[0]
        assert parse_span["parent_span_id"] == query_span["span_id"]
        assert parse_span["description"] == "parsing"

        validate_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_VALIDATE
        ]
        assert len(validate_spans) == 1, "exactly one validate span expected"
        validate_span = validate_spans[0]
        assert validate_span["parent_span_id"] == query_span["span_id"]
        assert validate_span["description"] == "validation"

        resolve_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_RESOLVE
        ]
        assert len(resolve_spans) == 1, "exactly one resolve span expected"
        resolve_span = resolve_spans[0]
        assert resolve_span["parent_span_id"] == query_span["span_id"]
        assert resolve_span["description"] == "resolving Query.hello"
        assert resolve_span["data"] == ApproxDict(
            {
                "graphql.field_name": "hello",
                "graphql.parent_type": "Query",
                "graphql.field_path": "Query.hello",
                "graphql.path": "hello",
            }
        )


@parameterize_strawberry_test
@pytest.mark.parametrize(
    "send_default_pii",
    [True, False],
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_transaction_no_operation_name(
    request,
    sentry_init,
    capture_events,
    capture_items,
    client_factory,
    async_execution,
    framework_integrations,
    send_default_pii,
    span_streaming,
):
    sentry_init(
        integrations=[
            StrawberryIntegration(async_execution=async_execution),
        ]
        + framework_integrations,
        traces_sample_rate=1,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    schema = strawberry.Schema(Query)

    client_factory = request.getfixturevalue(client_factory)
    client = client_factory(schema)

    query = "{ hello }"
    client.post("/graphql", json={"query": query})

    if span_streaming:
        sentry_sdk.flush()
        spans = [i.payload for i in items]

        assert len(spans) == 5
        parse_span, validate_span, resolve_span, query_span, segment = spans

        assert segment["is_segment"] is True
        if async_execution:
            assert segment["name"] == "/graphql"
        else:
            assert segment["name"] == "graphql_view"

        assert query_span["attributes"]["sentry.op"] == OP.GRAPHQL_QUERY
        assert query_span["name"] == "query"
        assert query_span["attributes"]["graphql.operation.type"] == "query"
        assert "graphql.operation.name" not in query_span["attributes"]

        if send_default_pii is True:
            assert query_span["attributes"]["graphql.document"] == query
        else:
            assert "graphql.document" not in query_span["attributes"]

        assert parse_span["attributes"]["sentry.op"] == OP.GRAPHQL_PARSE
        assert parse_span["name"] == "parsing"
        assert parse_span["parent_span_id"] == query_span["span_id"]

        assert validate_span["attributes"]["sentry.op"] == OP.GRAPHQL_VALIDATE
        assert validate_span["name"] == "validation"
        assert validate_span["parent_span_id"] == query_span["span_id"]

        assert resolve_span["attributes"]["sentry.op"] == OP.GRAPHQL_RESOLVE
        assert resolve_span["name"] == "resolving Query.hello"
        assert resolve_span["parent_span_id"] == query_span["span_id"]
    else:
        assert len(events) == 1
        (transaction_event,) = events

        if async_execution:
            assert transaction_event["transaction"] == "/graphql"
        else:
            assert transaction_event["transaction"] == "graphql_view"

        assert transaction_event["spans"]

        query_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_QUERY
        ]
        assert len(query_spans) == 1, "exactly one query span expected"
        query_span = query_spans[0]
        assert query_span["description"] == "query"
        assert query_span["data"]["graphql.operation.type"] == "query"
        assert query_span["data"]["graphql.operation.name"] is None
        assert query_span["data"]["graphql.resource_name"]

        if send_default_pii is True:
            assert query_span["data"]["graphql.document"] == query
        else:
            assert "graphql.document" not in query_span["data"]

        parse_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_PARSE
        ]
        assert len(parse_spans) == 1, "exactly one parse span expected"
        parse_span = parse_spans[0]
        assert parse_span["parent_span_id"] == query_span["span_id"]
        assert parse_span["description"] == "parsing"

        validate_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_VALIDATE
        ]
        assert len(validate_spans) == 1, "exactly one validate span expected"
        validate_span = validate_spans[0]
        assert validate_span["parent_span_id"] == query_span["span_id"]
        assert validate_span["description"] == "validation"

        resolve_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_RESOLVE
        ]
        assert len(resolve_spans) == 1, "exactly one resolve span expected"
        resolve_span = resolve_spans[0]
        assert resolve_span["parent_span_id"] == query_span["span_id"]
        assert resolve_span["description"] == "resolving Query.hello"
        assert resolve_span["data"] == ApproxDict(
            {
                "graphql.field_name": "hello",
                "graphql.parent_type": "Query",
                "graphql.field_path": "Query.hello",
                "graphql.path": "hello",
            }
        )


@parameterize_strawberry_test
@pytest.mark.parametrize(
    "send_default_pii",
    [True, False],
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_transaction_mutation(
    request,
    sentry_init,
    capture_events,
    capture_items,
    client_factory,
    async_execution,
    framework_integrations,
    send_default_pii,
    span_streaming,
):
    sentry_init(
        integrations=[
            StrawberryIntegration(async_execution=async_execution),
        ]
        + framework_integrations,
        traces_sample_rate=1,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    schema = strawberry.Schema(Query, mutation=Mutation)

    client_factory = request.getfixturevalue(client_factory)
    client = client_factory(schema)

    query = 'mutation Change { change(attribute: "something") }'
    client.post("/graphql", json={"query": query})

    if span_streaming:
        sentry_sdk.flush()
        spans = [i.payload for i in items]

        assert len(spans) == 5
        parse_span, validate_span, resolve_span, mutation_span, segment = spans

        assert segment["is_segment"] is True
        assert segment["name"] == "Change"
        assert segment["attributes"]["sentry.op"] == OP.GRAPHQL_MUTATION

        assert mutation_span["attributes"]["sentry.op"] == OP.GRAPHQL_MUTATION
        assert mutation_span["name"] == "mutation"
        assert mutation_span["attributes"]["graphql.operation.type"] == "mutation"
        assert "graphql.operation.name" not in mutation_span["attributes"]

        if send_default_pii is True:
            assert mutation_span["attributes"]["graphql.document"] == query
        else:
            assert "graphql.document" not in mutation_span["attributes"]

        assert parse_span["attributes"]["sentry.op"] == OP.GRAPHQL_PARSE
        assert parse_span["name"] == "parsing"
        assert parse_span["parent_span_id"] == mutation_span["span_id"]

        assert validate_span["attributes"]["sentry.op"] == OP.GRAPHQL_VALIDATE
        assert validate_span["name"] == "validation"
        assert validate_span["parent_span_id"] == mutation_span["span_id"]

        assert resolve_span["attributes"]["sentry.op"] == OP.GRAPHQL_RESOLVE
        assert resolve_span["name"] == "resolving Mutation.change"
        assert resolve_span["parent_span_id"] == mutation_span["span_id"]
    else:
        assert len(events) == 1
        (transaction_event,) = events

        assert transaction_event["transaction"] == "Change"
        assert transaction_event["contexts"]["trace"]["op"] == OP.GRAPHQL_MUTATION
        assert transaction_event["spans"]

        query_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_MUTATION
        ]
        assert len(query_spans) == 1, "exactly one mutation span expected"
        query_span = query_spans[0]
        assert query_span["description"] == "mutation"
        assert query_span["data"]["graphql.operation.type"] == "mutation"
        assert query_span["data"]["graphql.operation.name"] is None
        assert query_span["data"]["graphql.resource_name"]

        if send_default_pii is True:
            assert query_span["data"]["graphql.document"] == query
        else:
            assert "graphql.document" not in query_span["data"]

        parse_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_PARSE
        ]
        assert len(parse_spans) == 1, "exactly one parse span expected"
        parse_span = parse_spans[0]
        assert parse_span["parent_span_id"] == query_span["span_id"]
        assert parse_span["description"] == "parsing"

        validate_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_VALIDATE
        ]
        assert len(validate_spans) == 1, "exactly one validate span expected"
        validate_span = validate_spans[0]
        assert validate_span["parent_span_id"] == query_span["span_id"]
        assert validate_span["description"] == "validation"

        resolve_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.GRAPHQL_RESOLVE
        ]
        assert len(resolve_spans) == 1, "exactly one resolve span expected"
        resolve_span = resolve_spans[0]
        assert resolve_span["parent_span_id"] == query_span["span_id"]
        assert resolve_span["description"] == "resolving Mutation.change"
        assert resolve_span["data"] == ApproxDict(
            {
                "graphql.field_name": "change",
                "graphql.parent_type": "Mutation",
                "graphql.field_path": "Mutation.change",
                "graphql.path": "change",
            }
        )


@parameterize_strawberry_test
def test_handle_none_query_gracefully(
    request,
    sentry_init,
    capture_events,
    client_factory,
    async_execution,
    framework_integrations,
):
    sentry_init(
        integrations=[
            StrawberryIntegration(async_execution=async_execution),
        ]
        + framework_integrations,
    )
    events = capture_events()

    schema = strawberry.Schema(Query)

    client_factory = request.getfixturevalue(client_factory)
    client = client_factory(schema)

    client.post("/graphql", json={})

    assert len(events) == 0, "expected no events to be sent to Sentry"


@parameterize_strawberry_test
@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(
    request,
    sentry_init,
    capture_events,
    capture_items,
    client_factory,
    async_execution,
    framework_integrations,
    span_streaming,
):
    """
    Tests for OP.GRAPHQL_MUTATION, OP.GRAPHQL_PARSE, OP.GRAPHQL_VALIDATE, OP.GRAPHQL_RESOLVE,
    """
    sentry_init(
        integrations=[
            StrawberryIntegration(async_execution=async_execution),
        ]
        + framework_integrations,
        traces_sample_rate=1,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    schema = strawberry.Schema(Query, mutation=Mutation)

    client_factory = request.getfixturevalue(client_factory)
    client = client_factory(schema)

    query = 'mutation Change { change(attribute: "something") }'
    client.post("/graphql", json={"query": query})

    is_flask = "Flask" in str(framework_integrations[0])

    if span_streaming:
        sentry_sdk.flush()
        spans = [i.payload for i in items]

        assert len(spans) == 5
        parse_span, validate_span, resolve_span, mutation_span, segment = spans

        assert segment["is_segment"] is True
        if is_flask:
            assert segment["attributes"]["sentry.origin"] == "auto.http.flask"
        else:
            assert segment["attributes"]["sentry.origin"] == "auto.http.starlette"

        for span in (parse_span, validate_span, resolve_span, mutation_span):
            assert span["attributes"]["sentry.origin"] == "auto.graphql.strawberry"
    else:
        (event,) = events

        if is_flask:
            assert event["contexts"]["trace"]["origin"] == "auto.http.flask"
        else:
            assert event["contexts"]["trace"]["origin"] == "auto.http.starlette"

        for span in event["spans"]:
            if span["op"].startswith("graphql."):
                assert span["origin"] == "auto.graphql.strawberry"


@parameterize_strawberry_test
@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin2(
    request,
    sentry_init,
    capture_events,
    capture_items,
    client_factory,
    async_execution,
    framework_integrations,
    span_streaming,
):
    """
    Tests for OP.GRAPHQL_QUERY
    """
    sentry_init(
        integrations=[
            StrawberryIntegration(async_execution=async_execution),
        ]
        + framework_integrations,
        traces_sample_rate=1,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    schema = strawberry.Schema(Query, mutation=Mutation)

    client_factory = request.getfixturevalue(client_factory)
    client = client_factory(schema)

    query = "query GreetingQuery { hello }"
    client.post("/graphql", json={"query": query, "operationName": "GreetingQuery"})

    is_flask = "Flask" in str(framework_integrations[0])

    if span_streaming:
        sentry_sdk.flush()
        spans = [i.payload for i in items]

        assert len(spans) == 5
        parse_span, validate_span, resolve_span, query_span, segment = spans

        assert segment["is_segment"] is True
        if is_flask:
            assert segment["attributes"]["sentry.origin"] == "auto.http.flask"
        else:
            assert segment["attributes"]["sentry.origin"] == "auto.http.starlette"

        for span in (parse_span, validate_span, resolve_span, query_span):
            assert span["attributes"]["sentry.origin"] == "auto.graphql.strawberry"
    else:
        (event,) = events

        if is_flask:
            assert event["contexts"]["trace"]["origin"] == "auto.http.flask"
        else:
            assert event["contexts"]["trace"]["origin"] == "auto.http.starlette"

        for span in event["spans"]:
            if span["op"].startswith("graphql."):
                assert span["origin"] == "auto.graphql.strawberry"


@parameterize_strawberry_test
@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin3(
    request,
    sentry_init,
    capture_events,
    capture_items,
    client_factory,
    async_execution,
    framework_integrations,
    span_streaming,
):
    """
    Tests for OP.GRAPHQL_SUBSCRIPTION
    """
    sentry_init(
        integrations=[
            StrawberryIntegration(async_execution=async_execution),
        ]
        + framework_integrations,
        traces_sample_rate=1,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    schema = strawberry.Schema(Query, subscription=Subscription)

    client_factory = request.getfixturevalue(client_factory)
    client = client_factory(schema)

    query = "subscription { messageAdded { content } }"
    client.post("/graphql", json={"query": query})

    is_flask = "Flask" in str(framework_integrations[0])

    if span_streaming:
        sentry_sdk.flush()
        spans = [i.payload for i in items]

        assert len(spans) == 5
        parse_span, validate_span, resolve_span, subscription_span, segment = spans

        assert segment["is_segment"] is True
        if is_flask:
            assert segment["attributes"]["sentry.origin"] == "auto.http.flask"
        else:
            assert segment["attributes"]["sentry.origin"] == "auto.http.starlette"

        for span in (parse_span, validate_span, resolve_span, subscription_span):
            assert span["attributes"]["sentry.origin"] == "auto.graphql.strawberry"
    else:
        (event,) = events

        if is_flask:
            assert event["contexts"]["trace"]["origin"] == "auto.http.flask"
        else:
            assert event["contexts"]["trace"]["origin"] == "auto.http.starlette"

        for span in event["spans"]:
            if span["op"].startswith("graphql."):
                assert span["origin"] == "auto.graphql.strawberry"



import sys
from functools import wraps

import sentry_sdk
from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.integrations._wsgi_common import _filter_headers
from sentry_sdk.integrations.aws_lambda import _make_request_event_processor
from sentry_sdk.traces import (
    SpanStatus,
    StreamedSpan,
    get_current_span,
)
from sentry_sdk.tracing import TransactionSource
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    capture_internal_exceptions,
    event_from_exception,
    parse_version,
    reraise,
)

try:
    import chalice  # type: ignore
    from chalice import Chalice, ChaliceViewError
    from chalice import __version__ as CHALICE_VERSION
    from chalice.app import (  # type: ignore
        EventSourceHandler as ChaliceEventSourceHandler,
    )
except ImportError:
    raise DidNotEnable("Chalice is not installed")

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, TypeVar

    F = TypeVar("F", bound=Callable[..., Any])


class EventSourceHandler(ChaliceEventSourceHandler):  # type: ignore
    def __call__(self, event: "Any", context: "Any") -> "Any":
        client = sentry_sdk.get_client()

        with sentry_sdk.isolation_scope() as scope:
            with capture_internal_exceptions():
                configured_time = context.get_remaining_time_in_millis()
                scope.add_event_processor(
                    _make_request_event_processor(event, context, configured_time)
                )
            try:
                return ChaliceEventSourceHandler.__call__(self, event, context)
            except Exception:
                exc_info = sys.exc_info()
                event, hint = event_from_exception(
                    exc_info,
                    client_options=client.options,
                    mechanism={"type": "chalice", "handled": False},
                )
                sentry_sdk.capture_event(event, hint=hint)
                client.flush()
                reraise(*exc_info)


def _get_view_function_response(
    app: "Any", view_function: "F", function_args: "Any"
) -> "F":
    @wraps(view_function)
    def wrapped_view_function(**function_args: "Any") -> "Any":
        client = sentry_sdk.get_client()
        with sentry_sdk.isolation_scope() as scope:
            with capture_internal_exceptions():
                configured_time = app.lambda_context.get_remaining_time_in_millis()
                scope.add_event_processor(
                    _make_request_event_processor(
                        app.current_request.to_dict(),
                        app.lambda_context,
                        configured_time,
                    )
                )

            if has_span_streaming_enabled(client.options):
                current_span = get_current_span()
                segment = None
                if type(current_span) is StreamedSpan:
                    # A segment already exists (created by the AWS Lambda
                    # integration), so decorate it with Chalice attributes
                    # The AWS Lambda integration owns the span lifecycle
                    # (end + flush), but Chalice converts unhandled view exceptions
                    # into 500 responses, so the error must be captured here.
                    request_dict = app.current_request.to_dict()
                    headers = request_dict.get("headers", {})

                    header_attrs: "Dict[str, Any]" = {}
                    for header, value in _filter_headers(
                        headers, use_annotated_value=False
                    ).items():
                        header_attrs[f"http.request.header.{header.lower()}"] = value

                    additional_attrs: "Dict[str, Any]" = {}
                    if "method" in request_dict:
                        additional_attrs["http.request.method"] = request_dict["method"]

                    attributes = {
                        "sentry.origin": ChaliceIntegration.origin,
                        **header_attrs,
                        **additional_attrs,
                    }

                    segment = current_span._segment
                    segment.set_attributes(attributes)

                try:
                    return view_function(**function_args)
                except Exception as exc:
                    if isinstance(exc, ChaliceViewError):
                        raise
                    exc_info = sys.exc_info()
                    if segment:
                        segment.status = SpanStatus.ERROR.value
                    sentry_event, hint = event_from_exception(
                        exc_info,
                        client_options=client.options,
                        mechanism={"type": "chalice", "handled": False},
                    )
                    sentry_sdk.capture_event(sentry_event, hint=hint)
                    if segment is None:
                        client.flush()
                    raise
            else:
                scope.set_transaction_name(
                    app.lambda_context.function_name,
                    source=TransactionSource.COMPONENT,
                )
                try:
                    return view_function(**function_args)
                except Exception as exc:
                    if isinstance(exc, ChaliceViewError):
                        raise
                    exc_info = sys.exc_info()
                    sentry_event, hint = event_from_exception(
                        exc_info,
                        client_options=client.options,
                        mechanism={"type": "chalice", "handled": False},
                    )
                    sentry_sdk.capture_event(sentry_event, hint=hint)
                    client.flush()
                    raise

    return wrapped_view_function  # type: ignore


class ChaliceIntegration(Integration):
    identifier = "chalice"
    origin = f"auto.function.{identifier}"

    @staticmethod
    def setup_once() -> None:
        version = parse_version(CHALICE_VERSION)

        if version is None:
            raise DidNotEnable("Unparsable Chalice version: {}".format(CHALICE_VERSION))

        if version < (1, 20):
            old_get_view_function_response = Chalice._get_view_function_response
        else:
            from chalice.app import RestAPIEventHandler

            old_get_view_function_response = (
                RestAPIEventHandler._get_view_function_response
            )

        def sentry_event_response(
            app: "Any", view_function: "F", function_args: "Dict[str, Any]"
        ) -> "Any":
            wrapped_view_function = _get_view_function_response(
                app, view_function, function_args
            )

            return old_get_view_function_response(
                app, wrapped_view_function, function_args
            )

        if version < (1, 20):
            Chalice._get_view_function_response = sentry_event_response
        else:
            RestAPIEventHandler._get_view_function_response = sentry_event_response
        # for everything else (like events)
        chalice.app.EventSourceHandler = EventSourceHandler



import asyncio
import datetime
import json
import os
from contextlib import suppress
from unittest import mock

import pytest
from aiohttp import web
from aiohttp.client import ServerDisconnectedError
from aiohttp.web_exceptions import (
    HTTPBadRequest,
    HTTPInternalServerError,
    HTTPNetworkAuthenticationRequired,
    HTTPNotFound,
    HTTPUnavailableForLegalReasons,
)
from aiohttp.web_request import Request

import sentry_sdk
from sentry_sdk import capture_message, start_transaction
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations.aiohttp import (
    AioHttpIntegration,
    create_trace_config,
)
from sentry_sdk.utils import SENSITIVE_DATA_SUBSTITUTE
from tests.conftest import ApproxDict


@pytest.mark.asyncio
async def test_basic(sentry_init, aiohttp_client, capture_events):
    sentry_init(integrations=[AioHttpIntegration()])

    async def hello(request):
        1 / 0

    app = web.Application()
    app.router.add_get("/", hello)

    events = capture_events()

    client = await aiohttp_client(app)
    resp = await client.get("/")
    assert resp.status == 500

    (event,) = events

    assert (
        event["transaction"]
        == "tests.integrations.aiohttp.test_aiohttp.test_basic.<locals>.hello"
    )

    (exception,) = event["exception"]["values"]
    assert exception["type"] == "ZeroDivisionError"
    request = event["request"]
    host = request["headers"]["Host"]

    assert request["env"] == {"REMOTE_ADDR": "127.0.0.1"}
    assert request["method"] == "GET"
    assert request["query_string"] == ""
    assert request.get("data") is None
    assert request["url"] == "http://{host}/".format(host=host)
    assert request["headers"] == {
        "Accept": "*/*",
        "Accept-Encoding": mock.ANY,
        "Host": host,
        "User-Agent": request["headers"]["User-Agent"],
        "baggage": mock.ANY,
        "sentry-trace": mock.ANY,
    }


@pytest.mark.asyncio
async def test_post_body_not_read(sentry_init, aiohttp_client, capture_events):
    from sentry_sdk.integrations.aiohttp import BODY_NOT_READ_MESSAGE

    sentry_init(integrations=[AioHttpIntegration()])

    body = {"some": "value"}

    async def hello(request):
        1 / 0

    app = web.Application()
    app.router.add_post("/", hello)

    events = capture_events()

    client = await aiohttp_client(app)
    resp = await client.post("/", json=body)
    assert resp.status == 500

    (event,) = events
    (exception,) = event["exception"]["values"]
    assert exception["type"] == "ZeroDivisionError"
    request = event["request"]

    assert request["env"] == {"REMOTE_ADDR": "127.0.0.1"}
    assert request["method"] == "POST"
    assert request["data"] == BODY_NOT_READ_MESSAGE


@pytest.mark.asyncio
async def test_post_body_read(sentry_init, aiohttp_client, capture_events):
    sentry_init(integrations=[AioHttpIntegration()])

    body = {"some": "value"}

    async def hello(request):
        await request.json()
        1 / 0

    app = web.Application()
    app.router.add_post("/", hello)

    events = capture_events()

    client = await aiohttp_client(app)
    resp = await client.post("/", json=body)
    assert resp.status == 500

    (event,) = events
    (exception,) = event["exception"]["values"]
    assert exception["type"] == "ZeroDivisionError"
    request = event["request"]

    assert request["env"] == {"REMOTE_ADDR": "127.0.0.1"}
    assert request["method"] == "POST"
    assert request["data"] == json.dumps(body)


@pytest.mark.asyncio
async def test_403_not_captured(sentry_init, aiohttp_client, capture_events):
    sentry_init(integrations=[AioHttpIntegration()])

    async def hello(request):
        raise web.HTTPForbidden()

    app = web.Application()
    app.router.add_get("/", hello)

    events = capture_events()

    client = await aiohttp_client(app)
    resp = await client.get("/")
    assert resp.status == 403

    assert not events


@pytest.mark.asyncio
async def test_cancelled_error_not_captured(
    sentry_init, aiohttp_client, capture_events
):
    sentry_init(integrations=[AioHttpIntegration()])

    async def hello(request):
        raise asyncio.CancelledError()

    app = web.Application()
    app.router.add_get("/", hello)

    events = capture_events()
    client = await aiohttp_client(app)

    with suppress(ServerDisconnectedError):
        # Intended `aiohttp` interaction: server will disconnect if it
        # encounters `asyncio.CancelledError`
        await client.get("/")

    assert not events


@pytest.mark.asyncio
async def test_half_initialized(sentry_init, aiohttp_client, capture_events):
    sentry_init(integrations=[AioHttpIntegration()])
    sentry_init()

    async def hello(request):
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get("/", hello)

    events = capture_events()

    client = await aiohttp_client(app)
    resp = await client.get("/")
    assert resp.status == 200

    assert events == []


@pytest.mark.asyncio
async def test_tracing(sentry_init, aiohttp_client, capture_events):
    sentry_init(integrations=[AioHttpIntegration()], traces_sample_rate=1.0)

    async def hello(request):
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get("/", hello)

    events = capture_events()

    client = await aiohttp_client(app)
    resp = await client.get("/")
    assert resp.status == 200

    (event,) = events

    assert event["type"] == "transaction"
    assert (
        event["transaction"]
        == "tests.integrations.aiohttp.test_aiohttp.test_tracing.<locals>.hello"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "url,transaction_style,expected_transaction,expected_source",
    [
        (
            "/message",
            "handler_name",
            "tests.integrations.aiohttp.test_aiohttp.test_transaction_style.<locals>.hello",
            "component",
        ),
        (
            "/message",
            "method_and_path_pattern",
            "GET /{var}",
            "route",
        ),
    ],
)
async def test_transaction_style(
    sentry_init,
    aiohttp_client,
    capture_events,
    url,
    transaction_style,
    expected_transaction,
    expected_source,
):
    sentry_init(
        integrations=[AioHttpIntegration(transaction_style=transaction_style)],
        traces_sample_rate=1.0,
    )

    async def hello(request):
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get(r"/{var}", hello)

    events = capture_events()

    client = await aiohttp_client(app)
    resp = await client.get(url)
    assert resp.status == 200

    (event,) = events

    assert event["type"] == "transaction"
    assert event["transaction"] == expected_transaction
    assert event["transaction_info"] == {"source": expected_source}


@pytest.mark.tests_internal_exceptions
@pytest.mark.asyncio
async def test_tracing_unparseable_url(sentry_init, aiohttp_client, capture_events):
    sentry_init(integrations=[AioHttpIntegration()], traces_sample_rate=1.0)

    async def hello(request):
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get("/", hello)

    events = capture_events()

    client = await aiohttp_client(app)
    with mock.patch(
        "sentry_sdk.integrations.aiohttp.parse_url", side_effect=ValueError
    ):
        resp = await client.get("/")

    assert resp.status == 200

    (event,) = events

    assert event["type"] == "transaction"
    assert (
        event["transaction"]
        == "tests.integrations.aiohttp.test_aiohttp.test_tracing_unparseable_url.<locals>.hello"
    )


@pytest.mark.asyncio
async def test_traces_sampler_gets_request_object_in_sampling_context(
    sentry_init,
    aiohttp_client,
    DictionaryContaining,  # noqa: N803
    ObjectDescribedBy,  # noqa: N803
):
    traces_sampler = mock.Mock()
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sampler=traces_sampler,
    )

    async def kangaroo_handler(request):
        return web.Response(text="dogs are great")

    app = web.Application()
    app.router.add_get("/tricks/kangaroo", kangaroo_handler)

    client = await aiohttp_client(app)
    await client.get("/tricks/kangaroo")

    traces_sampler.assert_any_call(
        DictionaryContaining(
            {
                "aiohttp_request": ObjectDescribedBy(
                    type=Request, attrs={"method": "GET", "path": "/tricks/kangaroo"}
                )
            }
        )
    )


@pytest.mark.asyncio
async def test_has_trace_if_performance_enabled(
    sentry_init, aiohttp_client, capture_events
):
    sentry_init(integrations=[AioHttpIntegration()], traces_sample_rate=1.0)

    async def hello(request):
        capture_message("It's a good day to try dividing by 0")
        1 / 0

    app = web.Application()
    app.router.add_get("/", hello)

    events = capture_events()

    client = await aiohttp_client(app)
    resp = await client.get("/")
    assert resp.status == 500

    msg_event, error_event, transaction_event = events

    assert msg_event["contexts"]["trace"]
    assert "trace_id" in msg_event["contexts"]["trace"]

    assert error_event["contexts"]["trace"]
    assert "trace_id" in error_event["contexts"]["trace"]

    assert transaction_event["contexts"]["trace"]
    assert "trace_id" in transaction_event["contexts"]["trace"]

    assert (
        error_event["contexts"]["trace"]["trace_id"]
        == transaction_event["contexts"]["trace"]["trace_id"]
        == msg_event["contexts"]["trace"]["trace_id"]
    )


@pytest.mark.asyncio
async def test_has_trace_if_performance_disabled(
    sentry_init, aiohttp_client, capture_events
):
    sentry_init(integrations=[AioHttpIntegration()])

    async def hello(request):
        capture_message("It's a good day to try dividing by 0")
        1 / 0

    app = web.Application()
    app.router.add_get("/", hello)

    events = capture_events()

    client = await aiohttp_client(app)
    resp = await client.get("/")
    assert resp.status == 500

    msg_event, error_event = events

    assert msg_event["contexts"]["trace"]
    assert "trace_id" in msg_event["contexts"]["trace"]

    assert error_event["contexts"]["trace"]
    assert "trace_id" in error_event["contexts"]["trace"]

    assert (
        error_event["contexts"]["trace"]["trace_id"]
        == msg_event["contexts"]["trace"]["trace_id"]
    )


@pytest.mark.asyncio
async def test_trace_from_headers_if_performance_enabled(
    sentry_init, aiohttp_client, capture_events
):
    sentry_init(integrations=[AioHttpIntegration()], traces_sample_rate=1.0)

    async def hello(request):
        capture_message("It's a good day to try dividing by 0")
        1 / 0

    app = web.Application()
    app.router.add_get("/", hello)

    events = capture_events()

    # The aiohttp_client is instrumented so will generate the sentry-trace header and add request.
    # Get the sentry-trace header from the request so we can later compare with transaction events.
    client = await aiohttp_client(app)
    with start_transaction():
        # Headers are only added to the span if there is an active transaction
        resp = await client.get("/")

    sentry_trace_header = resp.request_info.headers.get("sentry-trace")
    trace_id = sentry_trace_header.split("-")[0]

    assert resp.status == 500

    # Last item is the custom transaction event wrapping `client.get("/")`
    msg_event, error_event, transaction_event, _ = events

    assert msg_event["contexts"]["trace"]
    assert "trace_id" in msg_event["contexts"]["trace"]

    assert error_event["contexts"]["trace"]
    assert "trace_id" in error_event["contexts"]["trace"]

    assert transaction_event["contexts"]["trace"]
    assert "trace_id" in transaction_event["contexts"]["trace"]

    assert msg_event["contexts"]["trace"]["trace_id"] == trace_id
    assert error_event["contexts"]["trace"]["trace_id"] == trace_id
    assert transaction_event["contexts"]["trace"]["trace_id"] == trace_id


@pytest.mark.asyncio
async def test_trace_from_headers_if_performance_disabled(
    sentry_init, aiohttp_client, capture_events
):
    sentry_init(integrations=[AioHttpIntegration()])

    async def hello(request):
        capture_message("It's a good day to try dividing by 0")
        1 / 0

    app = web.Application()
    app.router.add_get("/", hello)

    events = capture_events()

    # The aiohttp_client is instrumented so will generate the sentry-trace header and add request.
    # Get the sentry-trace header from the request so we can later compare with transaction events.
    client = await aiohttp_client(app)
    resp = await client.get("/")
    sentry_trace_header = resp.request_info.headers.get("sentry-trace")
    trace_id = sentry_trace_header.split("-")[0]

    assert resp.status == 500

    msg_event, error_event = events

    assert msg_event["contexts"]["trace"]
    assert "trace_id" in msg_event["contexts"]["trace"]

    assert error_event["contexts"]["trace"]
    assert "trace_id" in error_event["contexts"]["trace"]

    assert msg_event["contexts"]["trace"]["trace_id"] == trace_id
    assert error_event["contexts"]["trace"]["trace_id"] == trace_id


@pytest.mark.asyncio
async def test_crumb_capture(
    sentry_init, aiohttp_raw_server, aiohttp_client, capture_events
):
    def before_breadcrumb(crumb, hint):
        crumb["data"]["extra"] = "foo"
        return crumb

    sentry_init(
        integrations=[AioHttpIntegration()], before_breadcrumb=before_breadcrumb
    )

    async def handler(request):
        return web.Response(text="OK")

    raw_server = await aiohttp_raw_server(handler)

    with start_transaction():
        events = capture_events()

        client = await aiohttp_client(raw_server)
        resp = await client.get("/")
        assert resp.status == 200
        capture_message("Testing!")

        (event,) = events

        crumb = event["breadcrumbs"]["values"][0]
        assert crumb["type"] == "http"
        assert crumb["category"] == "httplib"
        assert crumb["data"] == ApproxDict(
            {
                "url": "http://127.0.0.1:{}/".format(raw_server.port),
                "http.fragment": "",
                "http.method": "GET",
                "http.query": "",
                "http.response.status_code": 200,
                "reason": "OK",
                "extra": "foo",
            }
        )


@pytest.mark.parametrize(
    "status_code,level",
    [
        (200, None),
        (301, None),
        (403, "warning"),
        (405, "warning"),
        (500, "error"),
    ],
)
@pytest.mark.asyncio
async def test_crumb_capture_client_error(
    sentry_init,
    aiohttp_raw_server,
    aiohttp_client,
    capture_events,
    status_code,
    level,
):
    sentry_init(integrations=[AioHttpIntegration()])

    async def handler(request):
        return web.Response(status=status_code)

    raw_server = await aiohttp_raw_server(handler)

    with start_transaction():
        events = capture_events()

        client = await aiohttp_client(raw_server)
        resp = await client.get("/")
        assert resp.status == status_code
        capture_message("Testing!")

        (event,) = events

        crumb = event["breadcrumbs"]["values"][0]
        assert crumb["type"] == "http"
        if level is None:
            assert "level" not in crumb
        else:
            assert crumb["level"] == level
        assert crumb["category"] == "httplib"
        assert crumb["data"] == ApproxDict(
            {
                "url": "http://127.0.0.1:{}/".format(raw_server.port),
                "http.fragment": "",
                "http.method": "GET",
                "http.query": "",
                "http.response.status_code": status_code,
            }
        )


@pytest.mark.asyncio
async def test_outgoing_trace_headers(sentry_init, aiohttp_raw_server, aiohttp_client):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
    )

    async def handler(request):
        return web.Response(text="OK")

    raw_server = await aiohttp_raw_server(handler)

    with start_transaction(
        name="/interactions/other-dogs/new-dog",
        op="greeting.sniff",
        # make trace_id difference between transactions
        trace_id="0123456789012345678901234567890",
    ) as transaction:
        client = await aiohttp_client(raw_server)
        resp = await client.get("/")
        request_span = transaction._span_recorder.spans[-1]

        assert resp.request_info.headers[
            "sentry-trace"
        ] == "{trace_id}-{parent_span_id}-{sampled}".format(
            trace_id=transaction.trace_id,
            parent_span_id=request_span.span_id,
            sampled=1,
        )


@pytest.mark.asyncio
async def test_outgoing_trace_headers_append_to_baggage(
    sentry_init, aiohttp_raw_server, aiohttp_client
):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        release="d08ebdb9309e1b004c6f52202de58a09c2268e42",
    )

    async def handler(request):
        return web.Response(text="OK")

    raw_server = await aiohttp_raw_server(handler)

    with mock.patch("sentry_sdk.tracing_utils.Random.randrange", return_value=500000):
        with start_transaction(
            name="/interactions/other-dogs/new-dog",
            op="greeting.sniff",
            trace_id="0123456789012345678901234567890",
        ):
            client = await aiohttp_client(raw_server)
            resp = await client.get("/", headers={"bagGage": "custom=value"})

            assert (
                resp.request_info.headers["baggage"]
                == "custom=value,sentry-trace_id=0123456789012345678901234567890,sentry-sample_rand=0.500000,sentry-environment=production,sentry-release=d08ebdb9309e1b004c6f52202de58a09c2268e42,sentry-transaction=/interactions/other-dogs/new-dog,sentry-sample_rate=1.0,sentry-sampled=true"
            )


@pytest.mark.asyncio
async def test_request_source_disabled(
    sentry_init,
    aiohttp_raw_server,
    aiohttp_client,
    capture_events,
):
    sentry_options = {
        "integrations": [AioHttpIntegration()],
        "traces_sample_rate": 1.0,
        "enable_http_request_source": False,
        "http_request_source_threshold_ms": 0,
    }

    sentry_init(**sentry_options)

    # server for making span request
    async def handler(request):
        return web.Response(text="OK")

    raw_server = await aiohttp_raw_server(handler)

    async def hello(request):
        span_client = await aiohttp_client(raw_server)
        await span_client.get("/")
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get(r"/", hello)

    events = capture_events()

    client = await aiohttp_client(app)
    await client.get("/")

    (event,) = events

    span = event["spans"][-1]
    assert span["description"].startswith("GET")

    data = span.get("data", {})

    assert SPANDATA.CODE_LINENO not in data
    assert SPANDATA.CODE_NAMESPACE not in data
    assert SPANDATA.CODE_FILEPATH not in data
    assert SPANDATA.CODE_FUNCTION not in data


@pytest.mark.asyncio
@pytest.mark.parametrize("enable_http_request_source", [None, True])
async def test_request_source_enabled(
    sentry_init,
    aiohttp_raw_server,
    aiohttp_client,
    capture_events,
    enable_http_request_source,
):
    sentry_options = {
        "integrations": [AioHttpIntegration()],
        "traces_sample_rate": 1.0,
        "http_request_source_threshold_ms": 0,
    }
    if enable_http_request_source is not None:
        sentry_options["enable_http_request_source"] = enable_http_request_source

    sentry_init(**sentry_options)

    # server for making span request
    async def handler(request):
        return web.Response(text="OK")

    raw_server = await aiohttp_raw_server(handler)

    async def hello(request):
        span_client = await aiohttp_client(raw_server)
        await span_client.get("/")
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get(r"/", hello)

    events = capture_events()

    client = await aiohttp_client(app)
    await client.get("/")

    (event,) = events

    span = event["spans"][-1]
    assert span["description"].startswith("GET")

    data = span.get("data", {})

    assert SPANDATA.CODE_LINENO in data
    assert SPANDATA.CODE_NAMESPACE in data
    assert SPANDATA.CODE_FILEPATH in data
    assert SPANDATA.CODE_FUNCTION in data


@pytest.mark.asyncio
async def test_request_source(
    sentry_init, aiohttp_raw_server, aiohttp_client, capture_events
):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=0,
    )

    # server for making span request
    async def handler(request):
        return web.Response(text="OK")

    raw_server = await aiohttp_raw_server(handler)

    async def handler_with_outgoing_request(request):
        span_client = await aiohttp_client(raw_server)
        await span_client.get("/")
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get(r"/", handler_with_outgoing_request)

    events = capture_events()

    client = await aiohttp_client(app)
    await client.get("/")

    (event,) = events

    span = event["spans"][-1]
    assert span["description"].startswith("GET")

    data = span.get("data", {})

    assert SPANDATA.CODE_LINENO in data
    assert SPANDATA.CODE_NAMESPACE in data
    assert SPANDATA.CODE_FILEPATH in data
    assert SPANDATA.CODE_FUNCTION in data

    assert type(data.get(SPANDATA.CODE_LINENO)) == int
    assert data.get(SPANDATA.CODE_LINENO) > 0
    assert (
        data.get(SPANDATA.CODE_NAMESPACE) == "tests.integrations.aiohttp.test_aiohttp"
    )
    assert data.get(SPANDATA.CODE_FILEPATH).endswith(
        "tests/integrations/aiohttp/test_aiohttp.py"
    )

    is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
    assert is_relative_path

    assert data.get(SPANDATA.CODE_FUNCTION) == "handler_with_outgoing_request"


@pytest.mark.asyncio
async def test_request_source_with_module_in_search_path(
    sentry_init, aiohttp_raw_server, aiohttp_client, capture_events
):
    """
    Test that request source is relative to the path of the module it ran in
    """
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=0,
    )

    # server for making span request
    async def handler(request):
        return web.Response(text="OK")

    raw_server = await aiohttp_raw_server(handler)

    from aiohttp_helpers.helpers import get_request_with_client

    async def handler_with_outgoing_request(request):
        span_client = await aiohttp_client(raw_server)
        await get_request_with_client(span_client, "/")
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get(r"/", handler_with_outgoing_request)

    events = capture_events()

    client = await aiohttp_client(app)
    await client.get("/")

    (event,) = events

    span = event["spans"][-1]
    assert span["description"].startswith("GET")

    data = span.get("data", {})

    assert SPANDATA.CODE_LINENO in data
    assert SPANDATA.CODE_NAMESPACE in data
    assert SPANDATA.CODE_FILEPATH in data
    assert SPANDATA.CODE_FUNCTION in data

    assert type(data.get(SPANDATA.CODE_LINENO)) == int
    assert data.get(SPANDATA.CODE_LINENO) > 0
    assert data.get(SPANDATA.CODE_NAMESPACE) == "aiohttp_helpers.helpers"
    assert data.get(SPANDATA.CODE_FILEPATH) == "aiohttp_helpers/helpers.py"

    is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
    assert is_relative_path

    assert data.get(SPANDATA.CODE_FUNCTION) == "get_request_with_client"


@pytest.mark.asyncio
async def test_no_request_source_if_duration_too_short(
    sentry_init, aiohttp_raw_server, aiohttp_client, capture_events
):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=100,
    )

    # server for making span request
    async def handler(request):
        return web.Response(text="OK")

    raw_server = await aiohttp_raw_server(handler)

    async def handler_with_outgoing_request(request):
        span_client = await aiohttp_client(raw_server)
        await span_client.get("/")
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get(r"/", handler_with_outgoing_request)

    events = capture_events()

    def fake_create_trace_context(*args, **kwargs):
        trace_context = create_trace_config()

        async def overwrite_timestamps(session, trace_config_ctx, params):
            span = trace_config_ctx.span
            span.start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
            span.timestamp = datetime.datetime(2024, 1, 1, microsecond=99999)

        trace_context.on_request_end.insert(0, overwrite_timestamps)

        return trace_context

    with mock.patch(
        "sentry_sdk.integrations.aiohttp.create_trace_config",
        fake_create_trace_context,
    ):
        client = await aiohttp_client(app)
        await client.get("/")

    (event,) = events

    span = event["spans"][-1]
    assert span["description"].startswith("GET")

    data = span.get("data", {})

    assert SPANDATA.CODE_LINENO not in data
    assert SPANDATA.CODE_NAMESPACE not in data
    assert SPANDATA.CODE_FILEPATH not in data
    assert SPANDATA.CODE_FUNCTION not in data


@pytest.mark.asyncio
async def test_request_source_if_duration_over_threshold(
    sentry_init, aiohttp_raw_server, aiohttp_client, capture_events
):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=100,
    )

    # server for making span request
    async def handler(request):
        return web.Response(text="OK")

    raw_server = await aiohttp_raw_server(handler)

    async def handler_with_outgoing_request(request):
        span_client = await aiohttp_client(raw_server)
        await span_client.get("/")
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get(r"/", handler_with_outgoing_request)

    events = capture_events()

    def fake_create_trace_context(*args, **kwargs):
        trace_context = create_trace_config()

        async def overwrite_timestamps(session, trace_config_ctx, params):
            span = trace_config_ctx.span
            span.start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
            span.timestamp = datetime.datetime(2024, 1, 1, microsecond=100001)

        trace_context.on_request_end.insert(0, overwrite_timestamps)

        return trace_context

    with mock.patch(
        "sentry_sdk.integrations.aiohttp.create_trace_config",
        fake_create_trace_context,
    ):
        client = await aiohttp_client(app)
        await client.get("/")

    (event,) = events

    span = event["spans"][-1]
    assert span["description"].startswith("GET")

    data = span.get("data", {})

    assert SPANDATA.CODE_LINENO in data
    assert SPANDATA.CODE_NAMESPACE in data
    assert SPANDATA.CODE_FILEPATH in data
    assert SPANDATA.CODE_FUNCTION in data

    assert type(data.get(SPANDATA.CODE_LINENO)) == int
    assert data.get(SPANDATA.CODE_LINENO) > 0
    assert (
        data.get(SPANDATA.CODE_NAMESPACE) == "tests.integrations.aiohttp.test_aiohttp"
    )
    assert data.get(SPANDATA.CODE_FILEPATH).endswith(
        "tests/integrations/aiohttp/test_aiohttp.py"
    )

    is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
    assert is_relative_path

    assert data.get(SPANDATA.CODE_FUNCTION) == "handler_with_outgoing_request"


@pytest.mark.asyncio
async def test_span_origin(
    sentry_init,
    aiohttp_raw_server,
    aiohttp_client,
    capture_events,
):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
    )

    # server for making span request
    async def handler(request):
        return web.Response(text="OK")

    raw_server = await aiohttp_raw_server(handler)

    async def hello(request):
        span_client = await aiohttp_client(raw_server)
        await span_client.get("/")
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get(r"/", hello)

    events = capture_events()

    client = await aiohttp_client(app)
    await client.get("/")

    (event,) = events
    assert event["contexts"]["trace"]["origin"] == "auto.http.aiohttp"
    assert event["spans"][0]["origin"] == "auto.http.aiohttp"


@pytest.mark.parametrize(
    ("integration_kwargs", "exception_to_raise", "should_capture"),
    (
        ({}, None, False),
        ({}, HTTPBadRequest, False),
        (
            {},
            HTTPUnavailableForLegalReasons(None),
            False,
        ),  # Highest 4xx status code (451)
        ({}, HTTPInternalServerError, True),
        ({}, HTTPNetworkAuthenticationRequired, True),  # Highest 5xx status code (511)
        ({"failed_request_status_codes": set()}, HTTPInternalServerError, False),
        (
            {"failed_request_status_codes": set()},
            HTTPNetworkAuthenticationRequired,
            False,
        ),
        ({"failed_request_status_codes": {404, *range(500, 600)}}, HTTPNotFound, True),
        (
            {"failed_request_status_codes": {404, *range(500, 600)}},
            HTTPInternalServerError,
            True,
        ),
        (
            {"failed_request_status_codes": {404, *range(500, 600)}},
            HTTPBadRequest,
            False,
        ),
    ),
)
@pytest.mark.asyncio
async def test_failed_request_status_codes(
    sentry_init,
    aiohttp_client,
    capture_events,
    integration_kwargs,
    exception_to_raise,
    should_capture,
):
    sentry_init(integrations=[AioHttpIntegration(**integration_kwargs)])
    events = capture_events()

    async def handle(_):
        if exception_to_raise is not None:
            raise exception_to_raise
        else:
            return web.Response(status=200)

    app = web.Application()
    app.router.add_get("/", handle)

    client = await aiohttp_client(app)
    resp = await client.get("/")

    expected_status = (
        200 if exception_to_raise is None else exception_to_raise.status_code
    )
    assert resp.status == expected_status

    if should_capture:
        (event,) = events
        assert event["exception"]["values"][0]["type"] == exception_to_raise.__name__
    else:
        assert not events


@pytest.mark.asyncio
async def test_failed_request_status_codes_with_returned_status(
    sentry_init, aiohttp_client, capture_events
):
    """
    Returning a web.Response with a failed_request_status_code should not be reported to Sentry.
    """
    sentry_init(integrations=[AioHttpIntegration(failed_request_status_codes={500})])
    events = capture_events()

    async def handle(_):
        return web.Response(status=500)

    app = web.Application()
    app.router.add_get("/", handle)

    client = await aiohttp_client(app)
    resp = await client.get("/")

    assert resp.status == 500
    assert not events


@pytest.mark.asyncio
async def test_failed_request_status_codes_non_http_exception(
    sentry_init, aiohttp_client, capture_events
):
    """
    If an exception, which is not an instance of HTTPException, is raised, it should be captured, even if
    failed_request_status_codes is empty.
    """
    sentry_init(integrations=[AioHttpIntegration(failed_request_status_codes=set())])
    events = capture_events()

    async def handle(_):
        1 / 0

    app = web.Application()
    app.router.add_get("/", handle)

    client = await aiohttp_client(app)
    resp = await client.get("/")
    assert resp.status == 500

    (event,) = events
    assert event["exception"]["values"][0]["type"] == "ZeroDivisionError"


@pytest.mark.asyncio
@pytest.mark.parametrize("send_pii", [True, False])
async def test_tracing_span_streaming(
    sentry_init, aiohttp_client, capture_items, send_pii
):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=send_pii,
        trace_lifecycle="stream",
    )

    async def hello(request):
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get("/", hello)

    items = capture_items("span")

    client = await aiohttp_client(app)
    resp = await client.get("/")
    assert resp.status == 200

    sentry_sdk.flush()

    # The server-side http.server span is the only segment. The aiohttp_client
    # fixture's outgoing http.client span is suppressed because there is no
    # active span when the test client makes the request.
    assert len(items) == 1

    (server_span,) = [item.payload for item in items]

    assert server_span["is_segment"] is True
    assert (
        server_span["name"]
        == "tests.integrations.aiohttp.test_aiohttp.test_tracing_span_streaming.<locals>.hello"
    )
    assert server_span["attributes"]["sentry.op"] == "http.server"
    assert server_span["attributes"]["sentry.origin"] == "auto.http.aiohttp"
    assert server_span["attributes"]["http.response.status_code"] == 200
    assert server_span["attributes"]["sentry.segment.name.source"] == "component"
    assert server_span["status"] == "ok"
    # No query string on the request, so the attribute should be omitted.
    assert "url.query" not in server_span["attributes"]

    # Request attributes derived directly from the aiohttp request.
    assert server_span["attributes"]["http.request.method"] == "GET"

    if send_pii:
        assert "client.address" in server_span["attributes"]
        assert "user.ip_address" in server_span["attributes"]

        url_full = server_span["attributes"]["url.full"]
        assert url_full.startswith("http://127.0.0.1:")
        assert url_full.endswith("/")

        url_path = server_span["attributes"]["url.path"]
        assert url_path == "/"
    else:
        assert "url.full" not in server_span["attributes"]
        assert "url.path" not in server_span["attributes"]
        assert "url.query" not in server_span["attributes"]

        assert "client.address" not in server_span["attributes"]
        assert "user.ip_address" not in server_span["attributes"]

    # aiohttp's test client always sends a Host header; we assert it propagates
    # into the span attributes via _filter_headers.
    assert "http.request.header.host" in server_span["attributes"]


@pytest.mark.asyncio
async def test_sensitive_header_scrubbing_span_streaming(
    sentry_init, aiohttp_client, capture_items
):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    async def hello(request):
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get("/", hello)

    items = capture_items("span")

    client = await aiohttp_client(app)
    resp = await client.get(
        "/",
        headers={
            "Authorization": "Bearer secret-token",
            "X-Custom-Header": "passthrough",
        },
    )
    assert resp.status == 200

    sentry_sdk.flush()

    (server_span,) = [item.payload for item in items]

    # send_default_pii defaults to False, so _filter_headers substitutes
    # sensitive headers with SENSITIVE_DATA_SUBSTITUTE ("[Filtered]"). The
    # original token must not leak.
    assert (
        server_span["attributes"]["http.request.header.authorization"]
        == SENSITIVE_DATA_SUBSTITUTE
    )
    # Non-sensitive headers pass through untouched.
    assert (
        server_span["attributes"]["http.request.header.x-custom-header"]
        == "passthrough"
    )


@pytest.mark.asyncio
async def test_sensitive_header_passthrough_with_pii_span_streaming(
    sentry_init, aiohttp_client, capture_items
):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=True,
        trace_lifecycle="stream",
    )

    async def hello(request):
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get("/", hello)

    items = capture_items("span")

    client = await aiohttp_client(app)
    await client.get("/", headers={"Authorization": "Bearer secret-token"})

    sentry_sdk.flush()

    (server_span,) = [item.payload for item in items]

    # With send_default_pii=True, _filter_headers is a no-op and the original
    # value reaches the span attribute.
    assert (
        server_span["attributes"]["http.request.header.authorization"]
        == "Bearer secret-token"
    )
    # client.address and user.ip_address is captured under send_default_pii=True.
    assert server_span["attributes"]["client.address"] == "127.0.0.1"
    assert server_span["attributes"]["user.ip_address"] == "127.0.0.1"


@pytest.mark.asyncio
@pytest.mark.parametrize("send_pii", [True, False])
async def test_url_query_attribute_span_streaming(
    sentry_init, aiohttp_client, capture_items, send_pii
):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=send_pii,
        trace_lifecycle="stream",
    )

    async def hello(request):
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get("/", hello)

    items = capture_items("span")

    client = await aiohttp_client(app)
    resp = await client.get("/?foo=bar&baz=qux")
    assert resp.status == 200

    sentry_sdk.flush()

    assert len(items) == 1
    (server_segment,) = [item.payload for item in items]

    if send_pii:
        assert server_segment["attributes"]["url.query"] == "foo=bar&baz=qux"
    else:
        assert "url.query" not in server_segment["attributes"]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "url,transaction_style,expected_name,expected_source",
    [
        (
            "/message",
            "handler_name",
            "tests.integrations.aiohttp.test_aiohttp."
            "test_transaction_style_span_streaming.<locals>.hello",
            "component",
        ),
        (
            "/message",
            "method_and_path_pattern",
            "GET /{var}",
            "route",
        ),
    ],
)
async def test_transaction_style_span_streaming(
    sentry_init,
    aiohttp_client,
    capture_items,
    url,
    transaction_style,
    expected_name,
    expected_source,
):
    sentry_init(
        integrations=[AioHttpIntegration(transaction_style=transaction_style)],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    async def hello(request):
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get(r"/{var}", hello)

    items = capture_items("span")

    client = await aiohttp_client(app)
    resp = await client.get(url)
    assert resp.status == 200

    sentry_sdk.flush()

    assert len(items) == 1
    (server_segment,) = [item.payload for item in items]

    assert server_segment["name"] == expected_name
    assert server_segment["is_segment"]
    assert server_segment["attributes"]["sentry.segment.name.source"] == expected_source


@pytest.mark.asyncio
async def test_server_error_span_streaming(sentry_init, aiohttp_client, capture_items):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    async def hello(request):
        1 / 0

    app = web.Application()
    app.router.add_get("/", hello)

    items = capture_items("event", "span")

    client = await aiohttp_client(app)
    resp = await client.get("/")
    assert resp.status == 500

    sentry_sdk.flush()

    # 1 error event + 1 span (server http.server)
    assert len(items) == 2

    error_event = items[0]
    assert error_event.type == "event"
    assert error_event.payload["exception"]["values"][0]["type"] == "ZeroDivisionError"

    server_span = items[1].payload

    # The integration's generic Exception path reraises without recording
    # http.response.status_code on the server span. StreamedSpan.__exit__
    # observes the propagating exception and sets status to "error".
    assert server_span["attributes"]["sentry.op"] == "http.server"
    assert "http.response.status_code" not in server_span["attributes"]
    assert server_span["status"] == "error"


@pytest.mark.asyncio
async def test_http_exception_span_streaming(
    sentry_init, aiohttp_client, capture_items
):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    async def hello(request):
        raise web.HTTPForbidden()

    app = web.Application()
    app.router.add_get("/", hello)

    items = capture_items("span")

    client = await aiohttp_client(app)
    resp = await client.get("/")
    assert resp.status == 403

    sentry_sdk.flush()

    assert len(items) == 1
    (server_span,) = [item.payload for item in items]

    assert server_span["attributes"]["sentry.op"] == "http.server"
    assert server_span["attributes"]["http.response.status_code"] == 403
    assert server_span["status"] == "error"


@pytest.mark.asyncio
async def test_http_exception_ok_status_not_overridden_span_streaming(
    sentry_init, aiohttp_client, capture_items
):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    async def hello(request):
        raise web.HTTPFound("https://example.com")

    app = web.Application()
    app.router.add_get("/", hello)

    items = capture_items("span")

    client = await aiohttp_client(app, server_kwargs={"skip_url_asserts": True})
    resp = await client.get("/", allow_redirects=False)
    assert resp.status == 302

    sentry_sdk.flush()

    assert len(items) == 1
    (server_span,) = [item.payload for item in items]

    assert server_span["attributes"]["sentry.op"] == "http.server"
    assert server_span["attributes"]["http.response.status_code"] == 302
    assert server_span["status"] == "ok"


@pytest.mark.asyncio
@pytest.mark.parametrize("send_pii", [True, False])
async def test_outgoing_client_span_span_streaming(
    sentry_init, aiohttp_raw_server, aiohttp_client, capture_items, send_pii
):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=send_pii,
        trace_lifecycle="stream",
    )

    async def handler(request):
        return web.Response(text="OK")

    raw_server = await aiohttp_raw_server(handler)

    async def hello(request):
        span_client = await aiohttp_client(raw_server)
        await span_client.get("/?foo=bar")
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get(r"/", hello)

    items = capture_items("span")

    client = await aiohttp_client(app)
    await client.get("/")

    sentry_sdk.flush()

    # 2 spans, finished inner-first:
    #   #0 inner http.client (server -> raw_server)
    #   #1 server http.server
    assert len(items) == 2

    inner_client_span = items[0].payload
    server_span = items[1].payload

    assert server_span["attributes"]["sentry.op"] == "http.server"

    assert inner_client_span["is_segment"] is False
    assert inner_client_span["name"].startswith("GET ")
    assert inner_client_span["attributes"]["sentry.op"] == "http.client"
    assert inner_client_span["attributes"]["sentry.origin"] == "auto.http.aiohttp"
    assert inner_client_span["attributes"]["http.request.method"] == "GET"
    assert inner_client_span["attributes"]["http.response.status_code"] == 200
    assert inner_client_span["status"] == "ok"

    if send_pii:
        assert inner_client_span["attributes"]["url.query"] == "foo=bar"

        url_full = inner_client_span["attributes"]["url.full"]

        # parse_url() splits the URL — url.full is the base URL only, with the
        # query string captured separately on url.query.
        assert url_full.startswith("http://127.0.0.1:")
        assert url_full.endswith("/")

        assert inner_client_span["attributes"]["url.path"] == "/"


@pytest.mark.asyncio
async def test_outgoing_trace_headers_span_streaming(
    sentry_init, aiohttp_raw_server, aiohttp_client, capture_items
):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    async def handler(request):
        return web.Response(text="OK")

    raw_server = await aiohttp_raw_server(handler)

    items = capture_items("span")

    client = await aiohttp_client(raw_server)
    resp = await client.get("/")

    sentry_sdk.flush()

    # The outgoing http.client span is suppressed because there is no active
    # span when the test client makes the request.
    assert len(items) == 0

    # Even though no span is created, the trace propagation headers must still
    # be added to the outgoing request so the trace is not broken.
    assert "sentry-trace" in resp.request_info.headers
    assert "baggage" in resp.request_info.headers


@pytest.mark.asyncio
@pytest.mark.parametrize("send_default_pii", [True, False])
async def test_user_ip_address_on_all_spans(
    sentry_init, aiohttp_client, capture_items, send_default_pii
):
    sentry_init(
        integrations=[AioHttpIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream",
    )

    async def hello(request):
        with sentry_sdk.traces.start_span(name="child-span"):
            pass
        return web.Response(text="hello")

    app = web.Application()
    app.router.add_get("/", hello)

    items = capture_items("span")

    client = await aiohttp_client(app)
    await client.get("/")

    sentry_sdk.flush()

    child_span, server_span = [item.payload for item in items]

    assert server_span["attributes"]["sentry.segment.name.source"] == "component"
    assert "sentry.segment.name.source" not in child_span["attributes"]

    if send_default_pii:
        assert server_span["attributes"]["user.ip_address"] == "127.0.0.1"
        assert child_span["attributes"]["user.ip_address"] == "127.0.0.1"
    else:
        assert "user.ip_address" not in server_span["attributes"]
        assert "user.ip_address" not in child_span["attributes"]
